# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.auth import get_user_model
from django.forms import widgets, Media
from django.forms.utils import ErrorDict
from django.utils.translation import ugettext_lazy as _
from djng.forms import fields
from sass_processor.processor import sass_processor
from shop.forms.base import DialogForm, DialogModelForm, UniqueEmailValidationMixin
from shop.forms.widgets import CheckboxInput, RadioSelect, Select
from shop.models.address import ShippingAddressModel, BillingAddressModel
from shop.models.customer import CustomerModel
from shop.modifiers.pool import cart_modifiers_pool


class CustomerForm(DialogModelForm):
    scope_prefix = 'customer'
    legend = _("Customer's Details")

    email = fields.EmailField(label=_("Email address"))
    first_name = fields.CharField(label=_("First Name"))
    last_name = fields.CharField(label=_("Last Name"))

    class Meta:
        model = CustomerModel
        exclude = ['user', 'recognized', 'number', 'last_access']
        custom_fields = ['email', 'first_name', 'last_name']

    def __init__(self, initial=None, instance=None, *args, **kwargs):
        initial = dict(initial) if initial else {}
        assert instance is not None
        initial.update(dict((f, getattr(instance, f)) for f in self.Meta.custom_fields))
        super(CustomerForm, self).__init__(initial=initial, instance=instance, *args, **kwargs)

    @property
    def media(self):
        return Media(css={'all': [sass_processor('shop/css/customer.scss')]})

    def save(self, commit=True):
        for f in self.Meta.custom_fields:
            setattr(self.instance, f, self.cleaned_data[f])
        return super(CustomerForm, self).save(commit)

    @classmethod
    def form_factory(cls, request, data, cart):
        customer_form = cls(data=data, instance=request.customer)
        if customer_form.is_valid():
            customer_form.instance.recognize_as_registered(request, commit=False)
            customer_form.save()
        return customer_form


class GuestForm(UniqueEmailValidationMixin, DialogModelForm):
    scope_prefix = 'guest'
    form_name = 'customer_form'  # Override form name to reuse template `customer-form.html`
    legend = _("Customer's Email")

    email = fields.EmailField(label=_("Email address"))

    class Meta:
        model = get_user_model()  # since we only use the email field, use the User model directly
        fields = ['email']

    def __init__(self, initial=None, instance=None, *args, **kwargs):
        if isinstance(instance, CustomerModel):
            instance = instance.user
        super(GuestForm, self).__init__(initial=initial, instance=instance, *args, **kwargs)

    @classmethod
    def form_factory(cls, request, data, cart):
        customer_form = cls(data=data, instance=request.customer.user)
        if customer_form.is_valid():
            request.customer.recognize_as_guest(request, commit=False)
            customer_form.save()
        return customer_form


class AddressForm(DialogModelForm):
    # field to be superseeded by a select widget
    active_priority = fields.CharField(
        required=False,
        widget=widgets.HiddenInput(),
    )

    use_primary_address = fields.BooleanField(
        label="use primary address",  # label will be overridden by Shipping/Billing/AddressForm
        required=False,
        initial=True,
        widget=CheckboxInput(),
    )

    plugin_fields = ['plugin_id', 'plugin_order', 'use_primary_address']

    class Meta:
        exclude = ['customer', 'priority']

    def __init__(self, initial=None, instance=None, cart=None, *args, **kwargs):
        self.cart = cart
        self.multi_addr = kwargs.pop('multi_addr', False)
        self.allow_use_primary = kwargs.pop('allow_use_primary', False)
        self.populate_siblings_summary()
        if instance:
            initial = dict(initial or {}, active_priority=instance.priority)
            if instance.address_type == 'shipping':
                initial['use_primary_address'] = cart.shipping_address is None
            else:  # address_type == billing
                initial['use_primary_address'] = cart.billing_address is None
        super(AddressForm, self).__init__(initial=initial, instance=instance, *args, **kwargs)

    @property
    def media(self):
        return Media(css={'all': [sass_processor('shop/css/address.scss')]})

    @classmethod
    def get_model(cls):
        return cls.Meta.model

    @classmethod
    def form_factory(cls, request, data, cart):
        """
        From the given request, update the database model.
        If the form data is invalid, return an error dictionary to update the response.
        """
        # search for the associated address DB instance or create a new one
        current_address = cls.get_address(cart)
        try:
            active_priority = int(data.get('active_priority'))
        except (ValueError, TypeError):
            if data.get('use_primary_address'):
                active_priority = 'nop'
            else:
                active_priority = data.get('active_priority', 'add')
            active_address = cls.get_model().objects.get_fallback(customer=request.customer)
        else:
            filter_args = dict(customer=request.customer, priority=active_priority)
            active_address = cls.get_model().objects.filter(**filter_args).first()

        if active_priority == 'add':
            # Add a newly filled address for the given customer
            address_form = cls(data=data, cart=cart)
            if address_form.is_valid():
                # prevent adding the same address twice
                all_field_names = [f.name for f in cls.get_model()._meta.get_fields()]
                filter_args = dict((attr, val) for attr, val in address_form.data.items()
                                   if attr in all_field_names and val)
                filter_args.update(customer=request.customer)
                try:
                    existing_address = cls.get_model().objects.get(**filter_args)
                except cls.get_model().DoesNotExist:
                    next_address = address_form.save(commit=False)
                    if next_address:
                        next_address.customer = request.customer
                        next_address.priority = cls.get_model().objects.get_max_priority(request.customer) + 1
                        next_address.save()
                        address_form.data.update(active_priority=str(next_address.priority))
                    else:
                        address_form.data.update(active_priority='nop')
                    address_form.set_address(cart, next_address)
                else:
                    address_form.set_address(cart, existing_address)
                address_form.populate_siblings_summary()
        elif active_address is None and not data.get('use_primary_address'):
            # customer selected 'Add another address', hence create a new empty form
            initial = dict((key, val) for key, val in data.items() if key in cls.plugin_fields)
            address_form = cls(initial=initial)
            address_form.data.update(address_form.get_initial_data())
            address_form.data.update(active_priority='add')
        elif current_address == active_address:
            # an existing entity of AddressModel was edited
            address_form = cls(data=data, instance=active_address, cart=cart)
            if address_form.is_valid():
                next_address = address_form.save()
                address_form.set_address(cart, next_address)
        else:
            # an address with another priority was selected
            initial = dict(data)
            for attr in cls().get_initial_data().keys():
                if hasattr(active_address, attr):
                    initial.update({attr: getattr(active_address, attr)})
            initial.update(active_priority=str(active_address.priority))
            address_form = cls(data=initial, instance=current_address, cart=cart)
            address_form.set_address(cart, active_address)
        return address_form

    def populate_siblings_summary(self):
        """
        Build a list of value-labels to populate the address choosing element
        """
        self.siblings_summary = []
        if self.cart is not None:
            AddressModel = self.get_model()
            addresses = AddressModel.objects.filter(customer=self.cart.customer).order_by('priority')
            for number, addr in enumerate(addresses, 1):
                self.siblings_summary.append({
                    'value': str(addr.priority),
                    'label': "{}. {}".format(number, addr.as_text().strip().replace('\n', ' – '))
                })

    def full_clean(self):
        super(AddressForm, self).full_clean()
        if self.is_bound and self['use_primary_address'].value():
            # reset errors, since then the form is always regarded as valid
            self._errors = ErrorDict()

    def save(self, commit=True):
        if not self['use_primary_address'].value():
            return super(AddressForm, self).save(commit)

    def get_response_data(self):
        return dict(self.data, siblings_summary=self.siblings_summary)

    def as_div(self):
        # Intentionally rendered without field `use_primary_address`, this must be added
        # on top of the form template manually
        self.fields.pop('use_primary_address', None)
        return super(AddressForm, self).as_div()

    def as_text(self):
        bound_field = self['use_primary_address']
        if bound_field.value():
            return bound_field.field.widget.choice_label
        return super(AddressForm, self).as_text()


class ShippingAddressForm(AddressForm):
    scope_prefix = 'shipping_address'
    legend = _("Shipping Address")

    class Meta(AddressForm.Meta):
        model = ShippingAddressModel
        widgets = {
            'country': Select(attrs={'ng-change': 'updateSiblingAddress()'}),
        }

    def __init__(self, *args, **kwargs):
        super(ShippingAddressForm, self).__init__(*args, **kwargs)
        self.fields['use_primary_address'].label = _("Use billing address for shipping")
        self.fields['use_primary_address'].widget.choice_label = self.fields['use_primary_address'].label  # Django < 1.11

    @classmethod
    def get_address(cls, cart):
        return cart.shipping_address

    def set_address(self, cart, instance):
        cart.shipping_address = instance if not self['use_primary_address'].value() else None


class BillingAddressForm(AddressForm):
    scope_prefix = 'billing_address'
    legend = _("Billing Address")

    class Meta(AddressForm.Meta):
        model = BillingAddressModel

    def __init__(self, *args, **kwargs):
        super(BillingAddressForm, self).__init__(*args, **kwargs)
        self.fields['use_primary_address'].label = _("Use shipping address for billing")
        self.fields['use_primary_address'].widget.choice_label = self.fields['use_primary_address'].label  # Django < 1.11

    @classmethod
    def get_address(cls, cart):
        return cart.billing_address

    def set_address(self, cart, instance):
        cart.billing_address = instance if not self['use_primary_address'].value() else None


class PaymentMethodForm(DialogForm):
    scope_prefix = 'payment_method'

    payment_modifier = fields.ChoiceField(
        label=_("Payment Method"),
        widget=RadioSelect(attrs={'ng-change': 'updateMethod()'}),
    )

    def __init__(self, *args, **kwargs):
        choices = [m.get_choice() for m in cart_modifiers_pool.get_payment_modifiers()
                   if not m.is_disabled(kwargs['cart'])]
        self.base_fields['payment_modifier'].choices = choices
        if len(choices) == 1:
            # if there is only one shipping method available, always set it as default
            try:
                kwargs['initial']['payment_modifier'] = choices[0][0]
            except KeyError:
                pass
        super(PaymentMethodForm, self).__init__(*args, **kwargs)

    def has_choices(self):
        return len(self.base_fields['payment_modifier'].choices) > 0

    @classmethod
    def form_factory(cls, request, data, cart):
        cart.update(request)
        payment_method_form = cls(data=data, cart=cart)
        if payment_method_form.is_valid():
            payment_data = data.get('payment_data') or {}
            cart.extra.update(payment_method_form.cleaned_data, payment_extra_data=payment_data)
        return payment_method_form


class ShippingMethodForm(DialogForm):
    scope_prefix = 'shipping_method'

    shipping_modifier = fields.ChoiceField(
        label=_("Shipping Method"),
        widget=RadioSelect(attrs={'ng-change': 'updateMethod()'}),
    )

    def __init__(self, *args, **kwargs):
        choices = [m.get_choice() for m in cart_modifiers_pool.get_shipping_modifiers()
                   if not m.is_disabled(kwargs['cart'])]
        self.base_fields['shipping_modifier'].choices = choices
        if len(choices) == 1:
            # with only one choice, initialize with it
            try:
                kwargs['initial']['shipping_modifier'] = choices[0][0]
            except KeyError:
                pass
        super(ShippingMethodForm, self).__init__(*args, **kwargs)

    def has_choices(self):
        return len(self.base_fields['shipping_modifier'].choices) > 0

    @classmethod
    def form_factory(cls, request, data, cart):
        cart.update(request)
        shipping_method_form = cls(data=data, cart=cart)
        if shipping_method_form.is_valid():
            cart.extra.update(shipping_method_form.cleaned_data)
        return shipping_method_form


class ExtraAnnotationForm(DialogForm):
    scope_prefix = 'extra_annotation'

    annotation = fields.CharField(
        label=_("Extra annotation for this order"),
        required=False,
        widget=widgets.Textarea,
    )

    @classmethod
    def form_factory(cls, request, data, cart):
        extra_annotation_form = cls(data=data)
        if extra_annotation_form.is_valid():
            cart.extra.update(extra_annotation_form.cleaned_data)
        return extra_annotation_form


class AcceptConditionForm(DialogForm):
    scope_prefix = 'accept_condition'

    accept = fields.BooleanField(
        required=True,
        widget=CheckboxInput(),
    )

    def __init__(self, data=None, initial=None, *args, **kwargs):
        plugin_id = data and data.get('plugin_id') or initial and initial.get('plugin_id') or 'none'
        scope_prefix = '{0}.plugin_{1}'.format(self.scope_prefix, plugin_id)
        self.form_name = '{0}.plugin_{1}'.format(self.form_name, plugin_id)
        super(AcceptConditionForm, self).__init__(data=data, initial=initial,
                                                  scope_prefix=scope_prefix, *args, **kwargs)

    @classmethod
    def form_factory(cls, request, data, cart):
        data = data or {'accept': False}
        accept_form = cls(data=data)
        return accept_form



from __future__ import unicode_literals

from django.conf import settings
from django.template.loader import select_template
from django.urls import NoReverseMatch, reverse
from django.utils.translation import get_language_from_request
from cms.models.pagemodel import Page
from rest_framework.serializers import CharField, BooleanField
from rest_auth import serializers
from shop.conf import app_settings
from shop.forms.auth import PasswordResetRequestForm
from rest_auth.serializers import LoginSerializer as DefaultLoginSerializer


class LoginSerializer(DefaultLoginSerializer):
    stay_logged_in = BooleanField(required=False)


class PasswordResetRequestSerializer(serializers.PasswordResetSerializer):
    password_reset_form_class = PasswordResetRequestForm
    invalid_password_reset_confirm_url = '/cms-page_or_view_with__reverse_id=password-reset-confirm__does-not-exist/'

    def save(self):
        subject_template = select_template([
            '{}/email/password-reset-subject.txt'.format(app_settings.APP_LABEL),
            'shop/email/password-reset-subject.txt',
        ])
        body_text_template = select_template([
            '{}/email/password-reset-body.txt'.format(app_settings.APP_LABEL),
            'shop/email/password-reset-body.txt',
        ])
        body_html_template = select_template([
            '{}/email/password-reset-body.html'.format(app_settings.APP_LABEL),
            'shop/email/password-reset-body.html',
        ])
        try:
            page = Page.objects.select_related('node').get(reverse_id='password-reset-confirm', publisher_is_draft=False)
        except Page.DoesNotExist:
            try:
                password_reset_confirm_url = reverse('password-reset-confirm')
            except NoReverseMatch:
                password_reset_confirm_url = self.invalid_password_reset_confirm_url
        else:
            language = get_language_from_request(self.context['request'])
            password_reset_confirm_url = page.get_absolute_url(language)
        opts = {
            'use_https': self.context['request'].is_secure(),
            'from_email': getattr(settings, 'DEFAULT_FROM_EMAIL'),
            'request': self.context['request'],
            'subject_template_name': subject_template.template.name,
            'email_template_name': body_text_template.template.name,
            'html_email_template_name': body_html_template.template.name,
            'extra_email_context': {'password_reset_confirm_url': password_reset_confirm_url}
        }
        self.reset_form.save(**opts)


class PasswordResetConfirmSerializer(serializers.PasswordResetConfirmSerializer):
    new_password1 = CharField(min_length=6, max_length=128)
    new_password2 = CharField(min_length=6, max_length=128)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from formtools.wizard.views import normalize_name

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.forms import widgets
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from cms.utils.helpers import classproperty

from djng.forms import fields, NgModelFormMixin, NgFormValidationMixin
from djng.styling.bootstrap3.forms import Bootstrap3Form, Bootstrap3ModelForm


class DialogFormMixin(NgModelFormMixin, NgFormValidationMixin):
    required_css_class = 'djng-field-required'

    def __init__(self, *args, **kwargs):
        kwargs.pop('cart', None)  # cart object must be removed, otherwise underlying methods complain
        auto_name = self.form_name  ## .replace('_form', '')
        kwargs.setdefault('auto_id', '{}-%s'.format(auto_name))
        super(DialogFormMixin, self).__init__(*args, **kwargs)

    @classproperty
    def form_name(cls):
        return normalize_name(cls.__name__)

    def clean(self):
        cleaned_data = dict(super(DialogFormMixin, self).clean())
        cleaned_data.pop('plugin_id', None)
        if cleaned_data.pop('plugin_order', None) is None:
            msg = "Field 'plugin_order' is a hidden but required field in each form inheriting from DialogFormMixin"
            raise ValidationError(msg)
        return cleaned_data

    def as_text(self):
        """
        Dialog Forms rendered as summary just display their values instead of input fields.
        This is useful to render a summary of a previously filled out form.
        """
        try:
            return mark_safe(self.instance.as_text())
        except (AttributeError, TypeError):
            output = []
            for name in self.fields.keys():
                bound_field = self[name]
                value = bound_field.value()
                if bound_field.is_hidden:
                    continue
                if isinstance(value, (list, tuple)):
                    line = []
                    cast_to = type(tuple(bound_field.field.choices)[0][0])
                    for v in value:
                        try:
                            line.append(dict(bound_field.field.choices)[cast_to(v)])
                        except (AttributeError, KeyError):
                            pass
                    output.append(force_text(', '.join(line)))
                elif value:
                    try:
                        value = dict(bound_field.field.choices)[value]
                    except (AttributeError, KeyError):
                        pass
                    output.append(force_text(value))
            return mark_safe('\n'.join(output))

    def get_response_data(self):
        """
        Hook to respond with an updated version of the form data. This response then shall
        override the forms content.
        """


class DialogForm(DialogFormMixin, Bootstrap3Form):
    """
    Base class for all dialog forms used with a DialogFormPlugin.
    """
    label_css_classes = 'control-label font-weight-bold'

    plugin_id = fields.CharField(
        widget=widgets.HiddenInput,
        required=False,
    )

    plugin_order = fields.CharField(
        widget=widgets.HiddenInput,
    )


class DialogModelForm(DialogFormMixin, Bootstrap3ModelForm):
    """
    Base class for all dialog model forms used with a DialogFormPlugin.
    """
    plugin_id = fields.CharField(
        widget=widgets.HiddenInput,
        required=False,
    )

    plugin_order = fields.CharField(widget=widgets.HiddenInput)

    @cached_property
    def field_css_classes(self):
        css_classes = {'*': getattr(Bootstrap3ModelForm, 'field_css_classes')}
        for name, field in self.fields.items():
            if not field.widget.is_hidden:
                css_classes[name] = [css_classes['*']]
                css_classes[name].append('{}-{}'.format(self.scope_prefix, name))
        return css_classes


class UniqueEmailValidationMixin(object):
    """
    A mixin added to forms which have to validate for the uniqueness of email addresses.
    """
    def clean_email(self):
        if not self.cleaned_data['email']:
            raise ValidationError(_("Please provide a valid e-mail address"))
        # check for uniqueness of email address
        if get_user_model().objects.filter(is_active=True, email=self.cleaned_data['email']).exists():
            msg = _("A customer with the e-mail address '{email}' already exists.\n"
                    "If you have used this address previously, try to reset the password.")
            raise ValidationError(msg.format(**self.cleaned_data))
        return self.cleaned_data['email']



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rest_framework import serializers
from shop.serializers.bases import BaseCustomerSerializer


class CustomerSerializer(BaseCustomerSerializer):
    """
    If the chosen customer model is the default :class:`shop.models.defaults.Customer`, then this
    serializer shall be used.

    If another customer model is used, then add a customized ``CustomerSerializer`` to your project
    and point your configuration settings ``SHOP_CUSTOMER_SERIALIZER`` onto it.
    """
    salutation = serializers.CharField(source='get_salutation_display', read_only=True)

    class Meta(BaseCustomerSerializer.Meta):
        fields = BaseCustomerSerializer.Meta.fields + ['salutation']



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured
from django.db import transaction
from django.utils.module_loading import import_string

from rest_framework import status
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from cms.plugin_pool import plugin_pool

from shop.conf import app_settings
from shop.models.cart import CartModel
from shop.serializers.checkout import CheckoutSerializer
from shop.serializers.cart import CartSerializer
from shop.modifiers.pool import cart_modifiers_pool


class CheckoutViewSet(GenericViewSet):
    """
    View for our REST endpoint to communicate with the various forms used during the checkout.
    """
    serializer_label = 'checkout'
    serializer_class = CheckoutSerializer
    cart_serializer_class = CartSerializer

    def __init__(self, **kwargs):
        super(CheckoutViewSet, self).__init__(**kwargs)
        self.dialog_forms = set([import_string(fc) for fc in app_settings.SHOP_DIALOG_FORMS])
        try:
            from shop.cascade.plugin_base import DialogFormPluginBase
        except ImproperlyConfigured:
            # cmsplugins_cascade has not been installed
            pass
        else:
            # gather form classes from Cascade plugins for our checkout views
            for p in plugin_pool.get_all_plugins():
                if issubclass(p, DialogFormPluginBase):
                    if hasattr(p, 'form_classes'):
                        self.dialog_forms.update([import_string(fc) for fc in p.form_classes])
                    if hasattr(p, 'form_class'):
                        self.dialog_forms.add(import_string(p.form_class))

    @action(methods=['put'], detail=False, url_path='upload')
    def upload(self, request):
        """
        Use this REST endpoint to upload the payload of all forms used to setup the checkout
        dialogs. This method takes care to dispatch the uploaded payload to each corresponding
        form.
        """
        # sort posted form data by plugin order
        cart = CartModel.objects.get_from_request(request)
        dialog_data = []
        for form_class in self.dialog_forms:
            if form_class.scope_prefix in request.data:
                if 'plugin_order' in request.data[form_class.scope_prefix]:
                    dialog_data.append((form_class, request.data[form_class.scope_prefix]))
                else:
                    for data in request.data[form_class.scope_prefix].values():
                        dialog_data.append((form_class, data))
        dialog_data = sorted(dialog_data, key=lambda tpl: int(tpl[1]['plugin_order']))

        # save data, get text representation and collect potential errors
        errors, response_data, set_is_valid = {}, {}, True
        with transaction.atomic():
            for form_class, data in dialog_data:
                form = form_class.form_factory(request, data, cart)
                if form.is_valid():
                    # empty error dict forces revalidation by the client side validation
                    errors[form_class.form_name] = {}
                else:
                    errors[form_class.form_name] = form.errors
                    set_is_valid = False

                # by updating the response data, we can override the form's content
                update_data = form.get_response_data()
                if isinstance(update_data, dict):
                    response_data[form_class.form_name] = update_data

            # persist changes in cart
            if set_is_valid:
                cart.save()

        # add possible form errors for giving feedback to the customer
        if set_is_valid:
            return Response(response_data)
        else:
            return Response(errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)

    @action(methods=['get'], detail=False, url_path='digest')
    def digest(self, request):
        """
        Returns the summaries of the cart and various checkout forms to be rendered in non-editable fields.
        """
        cart = CartModel.objects.get_from_request(request)
        cart.update(request)
        context = self.get_serializer_context()
        checkout_serializer = self.serializer_class(cart, context=context, label=self.serializer_label)
        cart_serializer = self.cart_serializer_class(cart, context=context, label='cart')
        response_data = {
            'checkout_digest': checkout_serializer.data,
            'cart_summary': cart_serializer.data,
        }
        return Response(data=response_data)

    @action(methods=['post'], detail=False, url_path='purchase')
    def purchase(self, request):
        """
        This is the final step on converting a cart into an order object. It normally is used in
        combination with the plugin :class:`shop.cascade.checkout.ProceedButtonPlugin` to render
        a button labeled "Purchase Now".
        """
        cart = CartModel.objects.get_from_request(request)
        cart.update(request)
        cart.save()

        response_data = {}
        try:
            # Iterate over the registered modifiers, and search for the active payment service provider
            for modifier in cart_modifiers_pool.get_payment_modifiers():
                if modifier.is_active(cart.extra.get('payment_modifier')):
                    expression = modifier.payment_provider.get_payment_request(cart, request)
                    response_data.update(expression=expression)
                    break
        except ValidationError as err:
            response_data = {'purchasing_error_message': '. '.join(err.detail)}
            return Response(data=response_data, status=status.HTTP_422_UNPROCESSABLE_ENTITY)
        return Response(data=response_data)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db.models import SET_DEFAULT

from shop import deferred
from shop.models.address import BaseShippingAddress, BaseBillingAddress
from shop.models.cart import BaseCart


class Cart(BaseCart):
    """
    Default materialized model for BaseCart containing common fields
    """
    shipping_address = deferred.ForeignKey(
        BaseShippingAddress,
        null=True,
        default=None,
        related_name='+',
        on_delete=SET_DEFAULT,
    )

    billing_address = deferred.ForeignKey(
        BaseBillingAddress,
        null=True,
        default=None,
        related_name='+',
        on_delete=SET_DEFAULT,
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _, override

from cms.menu_bases import CMSAttachMenu
from menus.base import NavigationNode
from menus.menu_pool import menu_pool


class CatalogMenu(CMSAttachMenu):
    name = _("Catalog Menu")

    def get_nodes(self, request):
        try:
            if self.instance.publisher_is_draft:
                productpage_set = self.instance.publisher_public.productpage_set
            else:
                productpage_set = self.instance.productpage_set
        except AttributeError:
            return []
        nodes = []
        with override(request.LANGUAGE_CODE):
            for id, productpage in enumerate(productpage_set.all(), 1):
                node = NavigationNode(
                    title=productpage.product.product_name,
                    url=productpage.product.get_absolute_url(),
                    id=id,
                )
                if hasattr(productpage.product, 'slug'):
                    node.path = productpage.product.slug
                nodes.append(node)
        return nodes

menu_pool.register_menu(CatalogMenu)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf.urls import url
from django.contrib import admin
from django.core.urlresolvers import reverse
from django.db.models import Sum
from django.forms import models, ValidationError
from django.http import HttpResponse
from django.template.loader import select_template
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from shop.conf import app_settings
from shop.admin.order import OrderItemInline
from shop.models.order import OrderItemModel
from shop.models.delivery import DeliveryModel
from shop.modifiers.pool import cart_modifiers_pool
from shop.serializers.delivery import DeliverySerializer
from shop.serializers.order import OrderDetailSerializer


class OrderItemForm(models.ModelForm):
    """
    This form handles an ordered item, but adds a number field to modify the number of
    items to deliver.
    """
    class Meta:
        model = OrderItemModel
        exclude = ()

    def __init__(self, *args, **kwargs):
        if 'instance' in kwargs:
            kwargs.setdefault('initial', {})
            deliver_quantity = kwargs['instance'].quantity - self.get_delivered(kwargs['instance'])
            kwargs['initial'].update(deliver_quantity=deliver_quantity)
        else:
            deliver_quantity = None
        super(OrderItemForm, self).__init__(*args, **kwargs)
        if deliver_quantity == 0:
            self['deliver_quantity'].field.widget.attrs.update(readonly='readonly')

    @classmethod
    def get_delivered(cls, instance):
        """
        Returns the quantity already delivered for this order item.
        """
        aggr = instance.deliver_item.aggregate(delivered=Sum('quantity'))
        return aggr['delivered'] or 0

    def clean(self):
        cleaned_data = super(OrderItemForm, self).clean()
        if cleaned_data.get('deliver_quantity') is not None:
            if cleaned_data['deliver_quantity'] < 0:
                raise ValidationError(_("Only a positive number of items can be delivered"), code='invalid')
            if cleaned_data['deliver_quantity'] > self.instance.quantity - self.get_delivered(self.instance):
                raise ValidationError(_("The number of items to deliver exceeds the ordered quantity"), code='invalid')
        return cleaned_data

    def has_changed(self):
        """Force form to changed"""
        return True


class OrderItemInlineDelivery(OrderItemInline):
    def get_fields(self, request, obj=None):
        fields = list(super(OrderItemInlineDelivery, self).get_fields(request, obj))
        if obj:
            if obj.status == 'pick_goods' and obj.unfulfilled_items > 0:
                fields[1] += ('deliver_quantity', 'canceled',)
            else:
                fields[1] += ('get_delivered', 'show_ready',)
        return fields

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super(OrderItemInlineDelivery, self).get_readonly_fields(request, obj))
        if obj:
            if not (obj.status == 'pick_goods' and obj.unfulfilled_items > 0):
                readonly_fields.extend(['get_delivered', 'show_ready'])
        return readonly_fields

    def get_formset(self, request, obj=None, **kwargs):
        """
        Add field `quantity` to the form on the fly, using the same numeric type as `OrderItem.quantity`
        """
        labels = {'quantity': _("Deliver quantity")}
        attrs = models.fields_for_model(obj.items.model, fields=['quantity'], labels=labels)
        # rename to deliver_quantity, since quantity is already used
        attrs['deliver_quantity'] = attrs.pop('quantity')
        if obj.status == 'pick_goods' and obj.unfulfilled_items > 0:
            attrs['deliver_quantity'].widget.attrs.update(style='width: 50px;')
        else:
            attrs['deliver_quantity'].required = False
        form = type(str('OrderItemForm'), (OrderItemForm,), attrs)
        labels = {'canceled': _("Cancel this item")}
        kwargs.update(form=form, labels=labels)
        formset = super(OrderItemInlineDelivery, self).get_formset(request, obj, **kwargs)
        return formset

    def get_delivered(self, obj=None):
        return OrderItemForm.get_delivered(obj)
    get_delivered.short_description = _("Delivered quantity")

    def show_ready(self, obj=None):
        return not obj.canceled
    show_ready.boolean = True
    show_ready.short_description = _("Ready for delivery")


def get_shipping_choices():
    choices = [sm.get_choice() for sm in cart_modifiers_pool.get_shipping_modifiers()]
    return choices


class DeliveryForm(models.ModelForm):
    shipping_method = models.ChoiceField(
        label=_("Shipping by"),
        choices=get_shipping_choices,
    )

    class Meta:
        model = DeliveryModel
        exclude = []

    def has_changed(self):
        return True

    def clean_shipping_method(self):
        if not self.cleaned_data['shipping_method']:
            return self.instance.shipping_method
        return self.cleaned_data['shipping_method']


class DeliveryInline(admin.TabularInline):
    model = DeliveryModel
    form = DeliveryForm
    extra = 0
    fields = ['shipping_id', 'shipping_method', 'delivered_items', 'print_out', 'fulfilled_at', 'shipped_at']
    readonly_fields = ['delivered_items', 'print_out', 'fulfilled_at', 'shipped_at']

    def has_delete_permission(self, request, obj=None):
        return False

    def get_max_num(self, request, obj=None, **kwargs):
        qs = self.model.objects.filter(order=obj)
        return qs.count()

    def get_fields(self, request, obj=None):
        assert obj is not None, "An Order object can not be added through the Django-Admin"
        fields = list(super(DeliveryInline, self).get_fields(request, obj))
        if not obj.allow_partial_delivery:
            fields.remove('delivered_items')
        return fields

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super(DeliveryInline, self).get_readonly_fields(request, obj))
        if not app_settings.SHOP_OVERRIDE_SHIPPING_METHOD or obj.status == 'ready_for_delivery':
            readonly_fields.append('shipping_method')
        return readonly_fields

    def get_formset(self, request, obj=None, **kwargs):
        formset = super(DeliveryInline, self).get_formset(request, obj, **kwargs)
        if not app_settings.SHOP_OVERRIDE_SHIPPING_METHOD or obj.status == 'ready_for_delivery':
            # make readonly field optional
            formset.form.base_fields['shipping_method'].required = False
        return formset

    def delivered_items(self, obj):
        aggr = obj.items.aggregate(quantity=Sum('quantity'))
        aggr['quantity'] = aggr['quantity'] or 0
        aggr.update(items=obj.items.count())
        return '{quantity}/{items}'.format(**aggr)
    delivered_items.short_description = _("Quantity/Items")

    def print_out(self, obj):
        if obj.fulfilled_at is None:
            return ''
        link = reverse('admin:print_delivery_note', args=(obj.id,)), _("Delivery Note")
        return format_html(
            '<span class="object-tools"><a href="{0}" class="viewsitelink" target="_new">{1}</a></span>',
            *link)
    print_out.short_description = _("Print out")

    def fulfilled(self, obj):
        if obj.fulfilled_at:
            return timezone.localtime(obj.fulfilled_at).ctime()  # TODO: find the correct time format
        return _("Pending")
    fulfilled.short_description = _("Fulfilled at")


class DeliveryOrderAdminMixin(object):
    """
    Add this mixin to the class defining the OrderAdmin
    """
    def get_urls(self):
        my_urls = [
            url(r'^(?P<delivery_pk>\d+)/print_delivery_note/$',
                self.admin_site.admin_view(self.render_delivery_note),
                name='print_delivery_note'),
        ]
        my_urls.extend(super(DeliveryOrderAdminMixin, self).get_urls())
        return my_urls

    def render_delivery_note(self, request, delivery_pk=None):
        template = select_template([
            '{}/print/delivery-note.html'.format(app_settings.APP_LABEL.lower()),
            'shop/print/delivery-note.html'
        ])
        delivery = DeliveryModel.objects.get(pk=delivery_pk)
        context = {'request': request, 'render_label': 'print'}
        customer_serializer = app_settings.CUSTOMER_SERIALIZER(delivery.order.customer)
        order_serializer = OrderDetailSerializer(delivery.order, context=context)
        delivery_serializer = DeliverySerializer(delivery, context=context)
        content = template.render({
            'customer': customer_serializer.data,
            'order': order_serializer.data,
            'delivery': delivery_serializer.data,
            'object': delivery,
        })
        return HttpResponse(content)

    def get_inline_instances(self, request, obj=None):
        assert obj is not None, "An Order object can not be added through the Django-Admin"
        assert hasattr(obj, 'associate_with_delivery'), "Add 'shop.shipping.workflows.SimpleShippingWorkflowMixin' " \
            "(or a class inheriting from thereof) to SHOP_ORDER_WORKFLOWS."
        inline_instances = list(super(DeliveryOrderAdminMixin, self).get_inline_instances(request, obj))
        if obj.associate_with_delivery:
            if obj.allow_partial_delivery:
                # replace `OrderItemInline` by `OrderItemInlineDelivery` for that instance.
                inline_instances = [
                    OrderItemInlineDelivery(self.model, self.admin_site) if isinstance(instance, OrderItemInline) else instance
                    for instance in inline_instances
                ]
            inline_instances.append(DeliveryInline(self.model, self.admin_site))
        return inline_instances

    def save_related(self, request, form, formsets, change):
        super(DeliveryOrderAdminMixin, self).save_related(request, form, formsets, change)
        if form.instance.status == 'pack_goods' and 'status' in form.changed_data:
            orderitem_formset = [fs for fs in formsets if issubclass(fs.model, OrderItemModel)][0]
            form.instance.update_or_create_delivery(orderitem_formset.cleaned_data)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db.models import Sum
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django_fsm import transition
from shop.models.delivery import DeliveryModel, DeliveryItemModel


class SimpleShippingWorkflowMixin(object):
    """
    Workflow for simply marking the state of an Order while picking, packing and shipping items.
    It does not create a Delivery object.

    Add this class to ``settings.SHOP_ORDER_WORKFLOWS`` to mix it into the merchants Order model.
    It is mutual exclusive with :class:`shop.shipping.workflows.CommissionGoodsWorkflowMixin` or
    :class:`shop.shipping.workflows.PartialDeliveryWorkflowMixin`.

    It adds all the methods required for state transitions, while picking and packing
    the ordered goods for shipping.
    """
    TRANSITION_TARGETS = {
        'pick_goods': _("Picking goods"),
        'pack_goods': _("Packing goods"),
        'ship_goods': _("Prepare for shipping"),
        'ready_for_delivery': _("Ready for delivery"),
    }

    @property
    def associate_with_delivery(self):
        """
        :returns: ``True`` if this Order requires a delivery object.
        """
        return False

    @property
    def allow_partial_delivery(self):
        """
        :returns: ``True`` if partial item delivery is allowed.
        """
        return False

    @transition(field='status', source='payment_confirmed', target='pick_goods',
                custom=dict(admin=True, button_name=_("Pick the goods")))
    def pick_goods(self, by=None):
        """Change status to 'pick_goods'."""

    @transition(field='status', source='pick_goods', target='pack_goods',
                custom=dict(admin=True, button_name=_("Pack the goods")))
    def pack_goods(self, by=None):
        """Change status to 'pack_goods'."""

    @transition(field='status', source='pack_goods', target='ship_goods',
                custom=dict(admin=True, button_name=_("Prepare for shipping")))
    def ship_goods(self, by=None):
        """
        Ship the goods. This method implicitly invokes
        :method:`shop.shipping.modifiers.ShippingModifier.ship_the_goods(delivery)`
        """

    @transition(field='status', source='ship_goods', target='ready_for_delivery',
                custom=dict(auto=True))
    def prepare_for_delivery(self, by=None):
        """
        Put the parcel into the outgoing delivery.
        This method is invoked automatically by `ship_goods`.
        """

    def update_or_create_delivery(self, orderitem_data):
        """
        Hook to create a delivery object with items.
        """


class CommissionGoodsWorkflowMixin(SimpleShippingWorkflowMixin):
    """
    Workflow to commission all ordered items in one common Delivery.

    Add this class to ``settings.SHOP_ORDER_WORKFLOWS`` to mix it into the merchants Order model.
    It is mutual exclusive with :class:`shop.shipping.workflows.SimpleShippingWorkflowMixin` or
    :class:`shop.shipping.workflows.PartialDeliveryWorkflowMixin`.

    It adds all the methods required for state transitions, while picking and packing
    the ordered goods for shipping.
    """
    @property
    def associate_with_delivery(self):
        return True

    @transition(field='status', source='ship_goods', target='ready_for_delivery',
                custom=dict(auto=True))
    def prepare_for_delivery(self, by=None):
        """Put the parcel into the outgoing delivery."""

    def update_or_create_delivery(self, orderitem_data):
        """
        Update or create a Delivery object for all items of this Order object.
        """
        delivery, _ = DeliveryModel.objects.get_or_create(
            order=self,
            shipping_id__isnull=True,
            shipped_at__isnull=True,
            shipping_method=self.extra.get('shipping_modifier'),
            defaults={'fulfilled_at': timezone.now()}
        )
        for item in self.items.all():
            DeliveryItemModel.objects.create(
                delivery=delivery,
                item=item,
                quantity=item.quantity,
            )


class PartialDeliveryWorkflowMixin(CommissionGoodsWorkflowMixin):
    """
    Workflow to optionally commission ordered items partially.

    Add this class to ``settings.SHOP_ORDER_WORKFLOWS`` to mix it into the merchants Order model.
    It is mutual exclusive with :class:`shop.shipping.workflows.SimpleShippingWorkflowMixin` or
    :class:`shop.shipping.workflows.CommissionGoodsWorkflowMixin`.

    This mixin supports partial delivery, hence check that a materialized representation of the
    models :class:`shop.models.delivery.DeliveryModel` and :class:`shop.models.delivery.DeliveryItemModel`
    exists and is instantiated.

    Importing the classes :class:`shop.models.defaults.delivery.DeliveryModel` and
    :class:`shop.models.defaults.delivery_item.DeliveryItemModel` into the merchants
    ``models.py``, usually is enough. This adds all the methods required for state transitions,
    while picking, packing and shipping the ordered goods for delivery.
    """
    @property
    def allow_partial_delivery(self):
        return True

    @cached_property
    def unfulfilled_items(self):
        unfulfilled_items = 0
        for order_item in self.items.all():
            if not order_item.canceled:
                aggr = order_item.deliver_item.aggregate(delivered=Sum('quantity'))
                unfulfilled_items += order_item.quantity - (aggr['delivered'] or 0)
        return unfulfilled_items

    def ready_for_picking(self):
        return self.is_fully_paid() and self.unfulfilled_items > 0

    def ready_for_shipping(self):
        return self.delivery_set.filter(shipped_at__isnull=True).exists()

    @transition(field='status', source='*', target='pick_goods', conditions=[ready_for_picking],
                custom=dict(admin=True, button_name=_("Pick the goods")))
    def pick_goods(self, by=None):
        """Change status to 'pick_goods'."""

    @transition(field='status', source=['pick_goods'], target='pack_goods',
                custom=dict(admin=True, button_name=_("Pack the goods")))
    def pack_goods(self, by=None):
        """Prepare shipping object and change status to 'pack_goods'."""

    @transition(field='status', source='*', target='ship_goods', conditions=[ready_for_shipping],
                custom=dict(admin=True, button_name=_("Ship the goods")))
    def ship_goods(self, by=None):
        """Ship the goods."""

    @transition(field='status', source='ship_goods', target='ready_for_delivery',
                custom=dict(auto=True))
    def prepare_for_delivery(self, by=None):
        """Put the parcel into the outgoing delivery."""

    def update_or_create_delivery(self, orderitem_data):
        """
        Update or create a Delivery object and associate with selected ordered items.
        """
        delivery, _ = DeliveryModel.objects.get_or_create(
            order=self,
            shipping_id__isnull=True,
            shipped_at__isnull=True,
            shipping_method=self.extra.get('shipping_modifier'),
            defaults={'fulfilled_at': timezone.now()}
        )

        # create a DeliveryItem object for each ordered item to be shipped with this delivery
        for data in orderitem_data:
            if data['deliver_quantity'] > 0 and not data['canceled']:
                DeliveryItemModel.objects.create(
                    delivery=delivery,
                    item=data['id'],
                    quantity=data['deliver_quantity'],
                )
        if not delivery.items.exists():
            # since no OrderItem was added to this delivery, discard it
            delivery.delete()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from rest_framework import serializers
from shop.conf import app_settings


class SerializeFormAsTextField(serializers.SerializerMethodField):
    def __init__(self, form_class_name, **kwargs):
        try:
            self.form_class = import_string(app_settings.SHOP_CASCADE_FORMS[form_class_name])
        except ImportError:
            msg = "Can not import Form class. Please check your settings directive SHOP_CASCADE_FORMS['{}']."
            raise ImproperlyConfigured(msg.format(form_class_name))
        super(SerializeFormAsTextField, self).__init__(**kwargs)

    def to_representation(self, value):
        method = getattr(self.parent, self.method_name)
        try:
            return method(self.form_class, value)
        except AttributeError:
            return


class CheckoutSerializer(serializers.Serializer):
    """
    Serializer to digest a summary of data required for the checkout.
    """
    customer_tag = SerializeFormAsTextField('CustomerForm')
    shipping_address_tag = SerializeFormAsTextField('ShippingAddressForm')
    billing_address_tag = SerializeFormAsTextField('BillingAddressForm')
    shipping_method_tag = SerializeFormAsTextField('ShippingMethodForm')
    payment_method_tag = SerializeFormAsTextField('PaymentMethodForm')
    extra_annotation_tag = SerializeFormAsTextField('ExtraAnnotationForm')

    def get_customer_tag(self, form_class, cart):
        form = form_class(instance=cart.customer)
        return form.as_text()

    def get_shipping_address_tag(self, form_class, cart):
        form = form_class(instance=cart.shipping_address, cart=cart)
        return form.as_text()

    def get_billing_address_tag(self, form_class, cart):
        form = form_class(instance=cart.billing_address, cart=cart)
        return form.as_text()

    def get_shipping_method_tag(self, form_class, cart):
        form = form_class(initial=cart.extra, cart=cart)
        return form.as_text()

    def get_payment_method_tag(self, form_class, cart):
        form = form_class(initial=cart.extra, cart=cart)
        return form.as_text()

    def get_extra_annotation_tag(self, form_class, cart):
        form = form_class(initial=cart.extra, cart=cart)
        return form.as_text()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

try:
    import redis
except ImportError:
    redis = None
from django.conf import settings
from django.dispatch import Signal


customer_recognized = Signal(providing_args=['customer', 'request'])

if redis and hasattr(settings, 'SESSION_REDIS'):
    redis_con = dict((key, settings.SESSION_REDIS[key]) for key in ['host', 'port', 'db', 'socket_timeout'])
    pool = redis.ConnectionPool(**redis_con)
    redis_con = redis.Redis(connection_pool=pool)
else:
    redis_con = type(str('Redis'), (), {'publish': lambda *args: None})()


def email_queued():
    """
    If SESSION_REDIS is configured, inform a separately running worker engine, that
    emails are ready for delivery. Call this function every time an email has been
    handled over to the Post-Office.
    """
    redis_con.publish('django-SHOP', 'send_queued_mail')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rest_framework import serializers
from shop.conf import app_settings
from shop.serializers.bases import BaseOrderItemSerializer


class OrderItemSerializer(BaseOrderItemSerializer):
    summary = serializers.SerializerMethodField(
        help_text="Sub-serializer for fields to be shown in the product's summary.")

    class Meta(BaseOrderItemSerializer.Meta):
        fields = ['line_total', 'unit_price', 'product_code', 'quantity', 'summary', 'extra']

    def get_summary(self, order_item):
        label = self.context.get('render_label', 'order')
        serializer_class = app_settings.PRODUCT_SUMMARY_SERIALIZER
        serializer = serializer_class(order_item.product, context=self.context,
                                      read_only=True, label=label)
        return serializer.data



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from six import with_metaclass
from django.db import models
from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from shop import deferred
from shop.models.order import BaseOrder, BaseOrderItem, OrderItemModel
from shop.modifiers.pool import cart_modifiers_pool


@python_2_unicode_compatible
class BaseDelivery(with_metaclass(deferred.ForeignKeyBuilder, models.Model)):
    """
    Shipping provider to keep track on each delivery.
    """
    order = deferred.ForeignKey(BaseOrder)

    shipping_id = models.CharField(
        _("Shipping ID"),
        max_length=255,
        null=True,
        blank=True,
        help_text=_("The transaction processor's reference"),
    )

    fulfilled_at = models.DateTimeField(
        _("Fulfilled at"),
        null=True,
        blank=True,
        help_text=_("Timestamp of delivery fulfillment"),
    )

    shipped_at = models.DateTimeField(
        _("Shipped at"),
        null=True,
        blank=True,
        help_text=_("Timestamp of delivery shipment"),
    )

    shipping_method = models.CharField(
        _("Shipping method"),
        max_length=50,
        help_text=_("The shipping backend used to deliver items of this order"),
    )

    class Meta:
        abstract = True
        unique_together = ['shipping_method', 'shipping_id']
        get_latest_by = 'shipped_at'

    def __str__(self):
        return _("Delivery ID: {}").format(self.id)

    @classmethod
    def perform_model_check(cls):
        for field in OrderItemModel._meta.fields:
            if field.attname == 'canceled' and field.get_internal_type() == 'BooleanField':
                break
        else:
             msg = "Class `{}` must implement a `BooleanField` named `canceled`, if used in combination with a Delivery model."
             raise ImproperlyConfigured(msg.format(OrderItemModel.__name__))

    def clean(self):
        if self.order._fsm_requested_transition == ('status', 'ship_goods') and not self.shipped_at:
            shipping_modifier = cart_modifiers_pool.get_active_shipping_modifier(self.shipping_method)
            shipping_modifier.ship_the_goods(self)

    def get_number(self):
        """
        Hook to get the delivery number.
        A class inheriting from Order may transform this into a string which is better readable.
        """
        if self.order.allow_partial_delivery:
            for part, delivery in enumerate(self.order.delivery_set.all(), 1):
                if delivery.pk == self.pk:
                    return "{} / {}".format(self.order.get_number(), part)
        return self.order.get_number()

DeliveryModel = deferred.MaterializedModel(BaseDelivery)


class BaseDeliveryItem(with_metaclass(deferred.ForeignKeyBuilder, models.Model)):
    """
    Abstract base class to keep track on the delivered quantity for each ordered item. Since the
    quantity can be any numerical value, it has to be defined by the class implementing this model.
    """
    delivery = deferred.ForeignKey(
        BaseDelivery,
        verbose_name=_("Delivery"),
        related_name='items',
        help_text=_("Refer to the shipping provider used to ship this item"),
    )

    item = deferred.ForeignKey(
        BaseOrderItem,
        related_name='deliver_item',
        verbose_name=_("Ordered item"),
    )

    class Meta:
        abstract = True
        verbose_name = _("Deliver item")
        verbose_name_plural = _("Deliver items")

    @classmethod
    def perform_model_check(cls):
        for order_field in OrderItemModel._meta.fields:
            if order_field.attname == 'quantity':
                break
        else:
            msg = "Class `{}` must implement a field named `quantity`."
            raise ImproperlyConfigured(msg.format(OrderItemModel.__name__))
        for deliver_field in OrderItemModel._meta.fields:
            if deliver_field.attname == 'quantity':
                break
        else:
            msg = "Class `{}` must implement a field named `quantity`."
            raise ImproperlyConfigured(msg.format(cls.__name__))
        if order_field.get_internal_type() != deliver_field.get_internal_type():
            msg = "Field `{}.quantity` must be of one same type `{}.quantity`."
            raise ImproperlyConfigured(msg.format(cls.__name__, OrderItemModel.__name__))

DeliveryItemModel = deferred.MaterializedModel(BaseDeliveryItem)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from cmsplugin_cascade.bootstrap4.mixins import BootstrapUtilities


CASCADE_PLUGINS = getattr(settings, 'SHOP_CASCADE_PLUGINS',
    ['auth', 'breadcrumb', 'catalog', 'cart', 'checkout', 'extensions', 'order', 'processbar', 'search']
)


def set_defaults(config):
    config.setdefault('plugins_with_extra_mixins', {})
    config.setdefault('plugins_with_extra_render_templates', {})
    config['plugins_with_extra_mixins'].setdefault('ShopReorderButtonPlugin', BootstrapUtilities(
        BootstrapUtilities.margins, BootstrapUtilities.floats,
    ))
    config['plugins_with_extra_mixins'].setdefault('ShopCancelOrderButtonPlugin', BootstrapUtilities(
        BootstrapUtilities.margins, BootstrapUtilities.floats,
    ))
    config['plugins_with_extra_mixins'].setdefault('ProcessNextStepPlugin', BootstrapUtilities(
        BootstrapUtilities.margins, BootstrapUtilities.floats,
    ))
    config['plugins_with_extra_mixins'].setdefault('ShopProceedButton', BootstrapUtilities(
        BootstrapUtilities.margins, BootstrapUtilities.floats,
    ))
    config['plugins_with_extra_mixins'].setdefault('ShopLeftExtension', BootstrapUtilities(
        BootstrapUtilities.paddings,
    ))
    config['plugins_with_extra_mixins'].setdefault('ShopRightExtension', BootstrapUtilities(
        BootstrapUtilities.paddings,
    ))
    config['plugins_with_extra_mixins'].setdefault('ShopAddToCartPlugin', BootstrapUtilities(
        BootstrapUtilities.margins,
    ))
    config['plugins_with_extra_mixins'].setdefault('BootstrapButtonPlugin', BootstrapUtilities(
        BootstrapUtilities.floats,
    ))
    config['plugins_with_extra_render_templates'].setdefault('BootstrapButtonPlugin', [
        ('shop/button.html', _("Responsive Feedback")),
        ('cascade/bootstrap4/button.html', _("Default")),
    ])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import models
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _
from post_office.models import EmailTemplate
from filer.fields.file import FilerFileField
from shop.conf import app_settings
from shop.models.fields import ChoiceEnum, ChoiceEnumField


class Notify(ChoiceEnum):
    RECIPIENT = 0, _("Recipient")
    VENDOR = 1, _("Vendor")
    CUSTOMER = 2, _("Customer")
    NOBODY = 9, _("Nobody")


class Notification(models.Model):
    """
    A task executed on receiving a signal.
    """
    name = models.CharField(
        max_length=50,
        verbose_name=_("Name"),
    )

    transition_target = models.CharField(
        max_length=50,
        verbose_name=_("Event"),
    )

    notify = ChoiceEnumField(
        _("Whom to notify"),
        enum_type=Notify,
    )

    recipient = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Recipient"),
        null=True,
        limit_choices_to={'is_staff': True},
    )

    mail_template = models.ForeignKey(
        EmailTemplate,
        verbose_name=_("Template"),
        limit_choices_to=Q(language__isnull=True) | Q(language=''),
    )

    class Meta:
        app_label = 'shop'
        verbose_name = _("Notification")
        verbose_name_plural = _("Notifications")
        ordering = ['transition_target', 'recipient_id']

    def __str__(self):
        return self.name

    def get_recipient(self, order):
        if self.notify is Notify.RECIPIENT:
            return self.recipient.email
        if self.notify is Notify.CUSTOMER:
            return order.customer.email
        if self.notify is Notify.VENDOR:
            if hasattr(order, 'vendor'):
                return order.vendor.email
            return app_settings.SHOP_VENDOR_EMAIL


class NotificationAttachment(models.Model):
    notification = models.ForeignKey(Notification)
    attachment = FilerFileField(null=True, blank=True, related_name='email_attachment')

    class Meta:
        app_label = 'shop'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rest_framework import serializers
from rest_framework.fields import empty
from shop.models.cart import CartModel
from shop.rest.money import MoneyField


class AddToCartSerializer(serializers.Serializer):
    """
    By default, this serializer is used by the view class :class:`shop.views.catalog.AddToCartView`,
    which handles the communication from the "Add to Cart" dialog box.

    If a product has variations, which influence the fields in the "Add to Cart" dialog box, then
    this serializer shall be overridden by a customized implementation. Such a customized "*Add to
    Cart*" serializer has to be connected to the ``AddToCartView``. This usually is achieved in
    the projects ``urls.py`` by changing the catalog's routing to:
    ```
    urlpatterns = [
        ...
        url(r'^(?P<slug>[\w-]+)/add-to-cart', AddToCartView.as_view(
            serializer_class=CustomAddToCartSerializer,
        )),
        ...
    ]
    ```
    """
    quantity = serializers.IntegerField(default=1, min_value=1)
    unit_price = MoneyField(read_only=True)
    subtotal = MoneyField(read_only=True)
    product = serializers.IntegerField(read_only=True, help_text="The product's primary key")
    product_code = serializers.CharField(read_only=True, help_text="Exact product code of the cart item")
    extra = serializers.DictField(read_only=True, default={})
    is_in_cart = serializers.BooleanField(read_only=True, default=False)

    def __init__(self, instance=None, data=empty, **kwargs):
        context = kwargs.get('context', {})
        if 'product' in context:
            instance = self.get_instance(context, data, kwargs)
            if data == empty:
                quantity = self.fields['quantity'].default
            else:
                quantity = self.fields['quantity'].to_internal_value(data['quantity'])
            instance.setdefault('quantity', quantity)
            instance.setdefault('subtotal', instance['quantity'] * instance['unit_price'])
            super(AddToCartSerializer, self).__init__(instance, data, context=context)
        else:
            super(AddToCartSerializer, self).__init__(instance, data, **kwargs)

    def get_instance(self, context, data, extra_args):
        """
        Method to store the ordered products in the cart item instance.
        Remember to override this method, if the ``product_code`` is part of the
        variation rather than being part of the product itself.
        """
        product = context['product']
        try:
            cart = CartModel.objects.get_from_request(context['request'])
        except CartModel.DoesNotExist:
            cart = None
        extra = data.get('extra', {}) if data is not empty else {}
        return {
            'product': product.id,
            'product_code': product.product_code,
            'unit_price': product.get_price(context['request']),
            'is_in_cart': bool(product.is_in_cart(cart)),
            'extra': extra,
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from cms.models.static_placeholder import StaticPlaceholder
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string
from cmsplugin_cascade.models import CascadeClipboard
from shop.management.utils import deserialize_to_placeholder


class MissingPage(CommandError):
    """
    Exception class indicating that a CMS page with a predefined ``reverse_id`` is missing.
    """


class MissingAppHook(CommandError):
    """
    Exception class indicating that a page misses the application.
    """


class MissingPlugin(CommandError):
    """
    Exception class indicating that a special plugin is missing or misconfigured on a given
    CMS page.
    """


class Command(BaseCommand):
    help = "Commands for Django-SHOP."

    def add_arguments(self, parser):
        parser.add_argument(
            'subcommand',
            help="./manage.py shop [customers|check-pages|review-settings]",
        )
        parser.add_argument(
            '--delete-expired',
            action='store_true',
            dest='delete_expired',
            help="Delete customers with expired sessions.",
        )
        parser.add_argument(
            '--add-missing',
            action='store_true',
            dest='add_missing',
            default=False,
            help="Use in combination with 'check-pages' to add missing mandatory pages.",
        )
        parser.add_argument(
            '--add-recommended',
            action='store_true',
            dest='add_recommended',
            default=False,
            help="Use in combination with 'check-pages' to add missing recommended pages.",
        )

    def handle(self, verbosity, subcommand, *args, **options):
        if subcommand == 'help':
            self.stdout.write("""
Usage:

./manage.py shop customers
    Show how many customers are registered, guests, anonymous or expired.            
    Use option --delete-expired to delete all customers with an expired session.            

./manage.py shop check-pages
    Iterate over all pages in the CMS and check, if they are properly configured.
    Use option --add-missing to add all missing mandatory pages for this shop.
    Use option --add-recommended to also add missing but recommended pages for this shop.

./manage.py shop review-settings
    Review all shop related settings and complain about missing- or mis-configurations.
""")
        elif subcommand == 'customers':
            self.delete_expired = options['delete_expired']
            self.customers()
        elif subcommand == 'check-pages':
            self.stdout.write("The following CMS pages must be adjusted:")
            self.add_recommended = options['add_recommended']
            self.add_mandatory = options['add_missing'] or self.add_recommended
            self.personal_pages = self.impersonal_pages = None
            if self.add_recommended:
                for k, msg in enumerate(self.create_recommended_pages(), 1):
                    self.stdout.write(" {}. {}".format(k, msg))
            for k, msg in enumerate(self.check_mandatory_pages(), 1):
                self.stdout.write(" {}. {}".format(k, msg))
        elif subcommand == 'review-settings':
            self.stdout.write("The following configuration settings must be fixed:")
            for k, msg in enumerate(self.review_settings(), 1):
                self.stdout.write(" {}. {}".format(k, msg))
        else:
            msg = "Unknown sub-command for shop. Use one of: customer check-pages review-settings"
            self.stderr.write(msg.format(subcommand))

    def customers(self):
        """
        Entry point for subcommand ``./manage.py shop customers``.
        """
        from shop.models.customer import CustomerModel

        data = dict(total=0, anonymous=0, active=0, staff=0, guests=0, registered=0, expired=0)
        for customer in CustomerModel.objects.iterator():
            data['total'] += 1
            if customer.user.is_active:
                data['active'] += 1
            if customer.user.is_staff:
                data['staff'] += 1
            if customer.is_registered:
                data['registered'] += 1
            elif customer.is_guest:
                data['guests'] += 1
            elif customer.is_anonymous:
                data['anonymous'] += 1
            if customer.is_expired:
                data['expired'] += 1
                if self.delete_expired and customer.orders.count() == 0:
                    customer.delete()
        msg = "Customers in this shop: total={total}, anonymous={anonymous}, expired={expired}, active={active}, guests={guests}, registered={registered}, staff={staff}."
        self.stdout.write(msg.format(**data))

    def create_recommended_pages(self):
        from cms.models.pagemodel import Page
        from cms.utils.i18n import get_public_languages

        default_language = get_public_languages()[0]

        # create the HOME page
        if Page.objects.public().filter(is_home=True).exists():
            yield "A home page exists already."
        else:
            page, created = self.get_or_create_page("Home", None, in_navigation=True)
            assert created is True
            try:
                clipboard = CascadeClipboard.objects.get(identifier='home')
            except CascadeClipboard.DoesNotExist:
                pass
            else:
                self.deserialize_to_placeholder(page, clipboard.data)
            try:
                clipboard = CascadeClipboard.objects.get(identifier='footer')
            except CascadeClipboard.DoesNotExist:
                pass
            else:
                static_placeholder = StaticPlaceholder.objects.create(code='Static Footer')
                deserialize_to_placeholder(static_placeholder.public, clipboard.data, default_language)
                deserialize_to_placeholder(static_placeholder.draft, clipboard.data, default_language)
            page.set_as_homepage()
            self.publish_in_all_languages(page)
            yield "Created recommended CMS home page."

        parent_page, created = self.get_or_create_page("Legal", None, reverse_id='shop-legal-pages', soft_root=True)
        if created:
            self.publish_in_all_languages(parent_page)
            yield "Created recommended CMS page 'Legal'."
        else:
            yield "Recommended CMS page 'Legal' exists already."

        page, created = self.get_or_create_page("Imprint", None, parent_page=parent_page, in_navigation=True)
        if created:
            self.publish_in_all_languages(page)
            yield "Created recommended CMS page 'Imprint'."
        else:
            yield "Recommended CMS page 'Imprint' exists already."

        page, created = self.get_or_create_page("Terms and Conditions", None, parent_page=parent_page, in_navigation=True)
        if created:
            self.publish_in_all_languages(page)
            yield "Created recommended CMS page 'Terms and Conditions'."
        else:
            yield "Recommended CMS page 'Terms and Conditions' exists already."

        page, created = self.get_or_create_page("Privacy Protection", None, parent_page=parent_page, in_navigation=True)
        if created:
            self.publish_in_all_languages(page)
            yield "Created recommended CMS page 'Privacy Protection'."
        else:
            yield "Recommended CMS page 'Privacy Protection' exists already."

        self.personal_pages, created = self.get_or_create_page("Personal Pages", None, reverse_id='shop-personal-pages', soft_root=True)
        if created:
            self.publish_in_all_languages(self.personal_pages)
            yield "Created recommended CMS page 'Personal Pages'."
        else:
            yield "Recommended CMS page 'Personal Pages' exists already."

        self.impersonal_pages, created = self.get_or_create_page("Join Us", None, reverse_id='shop-impersonal-pages', soft_root=True)
        if created:
            self.publish_in_all_languages(self.impersonal_pages)
            yield "Created recommended CMS page 'Join Us'."
        else:
            yield "Recommended CMS page 'Join Us' exists already."

    def check_mandatory_pages(self):
        """
        Entry point for subcommand ``./manage.py shop check-pages``.
        """
        from cms.models.pagemodel import Page
        from cms.models.pluginmodel import CMSPlugin
        from cms.utils.i18n import get_public_languages

        self._created_cms_pages = []
        default_language = get_public_languages()[0]

        # check for catalog pages
        apphook = self.get_installed_apphook('CatalogListCMSApp')
        catalog_pages = Page.objects.public().filter(application_urls=apphook.__class__.__name__)
        if not catalog_pages.exists():
            if self.add_mandatory:
                page, created = self.get_or_create_page("Catalog", 'CatalogListCMSApp', in_navigation=True)
                if created:
                    try:
                        clipboard = CascadeClipboard.objects.get(identifier='shop-list')
                        self.deserialize_to_placeholder(page, clipboard.data)
                    except CascadeClipboard.DoesNotExist:
                        leaf_plugin = self.create_page_structure(page)
                        self.add_plugin(leaf_plugin, 'ShopCatalogPlugin', {})
                    self.create_breadcrumb(page, {'render_type': 'catalog'})
                    self.publish_in_all_languages(page)
                    self.assign_all_products_to_page(page)
                    yield "Created CMS page titled 'Catalog'."
                else:
                    yield "CMS page 'Catalog' exists already."
            else:
                yield "There should be at least one published CMS page configured to use an Application inheriting from 'CatalogListCMSApp'."

        page_attributes = [
            # Menu Title, CMS-App-Hook or None, kwargs, Main Plugin, Plugin Context,
            (("Search", 'CatalogSearchCMSApp', {'reverse_id': 'shop-search-product'}), ('ShopSearchResultsPlugin', {}), {'render_type': 'catalog'}),
            (("Cart", None, {'reverse_id': 'shop-cart'}), ('ShopCartPlugin', {'render_type': 'editable'}), {'render_type': 'soft-root'}),
            (("Watch-List", None, {'reverse_id': 'shop-watch-list'}), ('ShopCartPlugin', {'render_type': 'watch'}), {'render_type': 'soft-root'}),
            (("Your Orders", 'OrderApp', {'reverse_id': 'shop-order', 'parent_page': self.personal_pages, 'in_navigation': True}), ('ShopOrderViewsPlugin', {}), None),
            (("Personal Details", None, {'reverse_id': 'shop-customer-details', 'parent_page': self.personal_pages, 'in_navigation': True}), ('CustomerFormPlugin', {}), None),
            (("Change Password", None, {'reverse_id': 'shop-password-change', 'parent_page': self.personal_pages, 'in_navigation': True}), ('ShopAuthenticationPlugin', {'form_type': 'password-change'}), None),
            (("Login", None, {'reverse_id': 'shop-login', 'parent_page': self.impersonal_pages, 'in_navigation': True}), ('ShopAuthenticationPlugin', {'form_type': 'login'}), None),
            (("Register Customer", None, {'reverse_id': 'shop-register-customer', 'parent_page': self.impersonal_pages, 'in_navigation': True}), ('ShopAuthenticationPlugin', {'form_type': 'register-user'}), None),
            (("Request Password Reset", None,  {'reverse_id': 'password-reset-request', 'parent_page': self.impersonal_pages, 'in_navigation': True}), ('ShopAuthenticationPlugin', {'form_type': 'password-reset-request'}), None),
            (("Confirm Password Reset", 'PasswordResetApp',  {'reverse_id': 'password-reset-confirm'}), ('ShopAuthenticationPlugin', {'form_type': 'password-reset-confirm'}), None),
            (("Payment Canceled", None, {'reverse_id': 'shop-cancel-payment'}), ('HeadingPlugin', {'tag_type': "h2", 'content': "Your payment has been canceled"}), None),
        ]
        for page_attrs, content_attrs, breadcrumb_glossary in page_attributes:
            try:
                page = self.check_page(*page_attrs[:2], **page_attrs[2])
                self.check_page_content(page, *content_attrs)
            except MissingPage as exc:
                if self.add_mandatory:
                    page, created = self.get_or_create_page(*page_attrs[:2], **page_attrs[2])
                    if created:
                        try:
                            clipboard = CascadeClipboard.objects.get(identifier=page.get_slug(default_language))
                            self.deserialize_to_placeholder(page, clipboard.data)
                        except CascadeClipboard.DoesNotExist:
                            leaf_plugin = self.create_page_structure(page)
                            self.add_plugin(leaf_plugin, *content_attrs)
                        if breadcrumb_glossary:
                            self.create_breadcrumb(page, breadcrumb_glossary)
                        self.publish_in_all_languages(page)
                        yield "Created mandatory CMS page titled '{0}'.".format(page.get_title(default_language))
                    else:
                        yield "Mandatory CMS page '{0}' exists already.".format(page.get_title(default_language))
                else:
                    yield str(exc)
            except CommandError as exc:
                yield str(exc)

        # the checkout page must be found through the purchase button
        for plugin in CMSPlugin.objects.filter(plugin_type='ShopProceedButton', language=default_language, placeholder__page__publisher_is_draft=False):
            link = plugin.get_bound_plugin().glossary.get('link')
            if isinstance(link, dict) and link.get('type') == 'PURCHASE_NOW':
                break
        else:
            if self.add_mandatory:
                page, created = self.get_or_create_page("Checkout", None)
                if created:
                    try:
                        clipboard = CascadeClipboard.objects.get(identifier='checkout')
                        self.deserialize_to_placeholder(page, clipboard.data)
                    except CascadeClipboard.DoesNotExist:
                        column_plugin = self.create_page_structure(page)
                        forms_plugin = self.add_plugin(column_plugin, 'ValidateSetOfFormsPlugin', {})
                        glossary = {'button_type': 'btn-success', 'link': {'type': 'PURCHASE_NOW'}, 'link_content': "Purchase Now"}
                        self.add_plugin(forms_plugin, 'ShopProceedButton', glossary)
                    self.publish_in_all_languages(page)
                    yield "Created CMS page titled 'Checkout'"
                else:
                    yield "CMS page titled 'Checkout' exists already."
            else:
                yield "There should be at least one published CMS page containing a 'Proceed Button Plugin' for purchasing the cart's content."

    @classmethod
    def get_installed_apphook(cls, base_apphook_name):
        from cms.apphook_pool import apphook_pool
        base_apphook = import_string('shop.cms_apphooks.' + base_apphook_name)

        for apphook, _ in apphook_pool.get_apphooks():
            apphook = apphook_pool.get_apphook(apphook)
            if isinstance(apphook, base_apphook):
                return apphook
        else:
            msg = "The project must register an AppHook inheriting from '{apphook_name}'"
            raise MissingAppHook(msg.format(apphook_name=base_apphook_name))

    def check_page(self, title, base_apphook_name, reverse_id=None, **kwargs):
        from cms.models.pagemodel import Page
        from cms.apphook_pool import apphook_pool

        page = Page.objects.public().filter(reverse_id=reverse_id).first()
        if not page:
            msg = "There should be a published CMS page with a reference ID: '{reverse_id}'."
            raise MissingPage(msg.format(reverse_id=reverse_id))

        if base_apphook_name:
            apphook = self.get_installed_apphook(base_apphook_name)
            if not page.application_urls or apphook_pool.get_apphook(page.application_urls) is not apphook:
                msg = "Page on URL '{url}' must be configured to use Application inheriting from '{app_hook}'."
                raise MissingAppHook(msg.format(url=page.get_absolute_url(), app_hook=base_apphook_name))

        return page

    def check_page_content(self, page, plugin_type, subset):
        from cms.plugin_pool import plugin_pool

        placeholder = page.placeholders.filter(slot='Main Content').first()
        if not placeholder:
            msg = "Page on URL '{url}' does not contain any plugin."
            raise MissingPlugin(msg.format(url=page.get_absolute_url()))

        plugin_name = plugin_pool.get_plugin(plugin_type).name
        for language in page.get_languages():
            plugin = placeholder.cmsplugin_set.filter(plugin_type=plugin_type, language=language).first()
            if not plugin:
                msg = "Page on URL '{url}' shall contain a plugin named '{plugin_name}'."
                raise MissingPlugin(msg.format(url=page.get_absolute_url(), plugin_name=plugin_name))

            glossary_items = plugin.get_bound_plugin().glossary.items()
            if not all(item in glossary_items for item in subset.items()):
                msg = "Plugin named '{plugin_name}' on page with URL '{url}' is misconfigured."
                raise MissingPlugin(msg.format(url=page.get_absolute_url(), plugin_name=plugin_name))

    def get_or_create_page(self, title, base_apphook_name, reverse_id=None, parent_page=None, in_navigation=False, soft_root=False):
        from cms.api import create_page
        from cms.models.pagemodel import Page
        from cms.utils.i18n import get_public_languages

        template = settings.CMS_TEMPLATES[0][0]
        apphook = self.get_installed_apphook(base_apphook_name) if base_apphook_name else None
        language = get_public_languages()[0]
        try:
            parent_node = parent_page.node if parent_page else None
            page = Page.objects.drafts().get(
                reverse_id=reverse_id,
                node__parent=parent_node,
                title_set__title=title,
                title_set__language=language,
            )
            created = False
        except Page.DoesNotExist:
            page = create_page(
                title,
                template,
                language,
                apphook=apphook,
                created_by="manage.py shop check-pages",
                in_navigation=in_navigation,
                soft_root=soft_root,
                parent=parent_page,
                reverse_id=reverse_id,
            )
            created = True
        return page, created

    def create_page_structure(self, page, slot='Main Content'):
        from cms.api import add_plugin
        from cms.utils.i18n import get_public_languages

        placeholder = page.placeholders.get(slot=slot)
        language = get_public_languages()[0]
        glossary = {
            'breakpoints': ['xs', 'sm', 'md', 'lg', 'xl'],
            'fluid': None,
        }
        container = add_plugin(placeholder, 'BootstrapContainerPlugin', language, glossary=glossary)
        row = add_plugin(placeholder, 'BootstrapRowPlugin', language, target=container)
        glossary = {
            'xs-column-width': 'col',
        }
        return add_plugin(placeholder, 'BootstrapColumnPlugin', language, target=row, glossary=glossary)

    def create_breadcrumb(self, page, glossary, slot='Breadcrumb'):
        from cms.api import add_plugin
        from cms.utils.i18n import get_public_languages

        placeholder = page.placeholders.get(slot=slot)
        language = get_public_languages()[0]
        return add_plugin(placeholder, 'BreadcrumbPlugin', language, glossary=glossary)

    def add_plugin(self, leaf_plugin, plugin_type, glossary):
        from cms.api import add_plugin
        if plugin_type:
            return add_plugin(leaf_plugin.placeholder, plugin_type, leaf_plugin.language, target=leaf_plugin, glossary=glossary)

    @classmethod
    def publish_in_all_languages(cls, page):
        from cms.api import copy_plugins_to_language, create_title
        from cms.utils.i18n import get_public_languages

        languages = get_public_languages()
        for language in languages[1:]:
            create_title(language, page.get_title(), page, menu_title=None)
            copy_plugins_to_language(page, languages[0], language)
        for language in languages:
            page.publish(language)

    def assign_all_products_to_page(self, page):
        from shop.models.product import ProductModel
        from shop.models.related import ProductPageModel

        for product in ProductModel.objects.all():
            ProductPageModel.objects.create(page=page, product=product)

    def review_settings(self):
        from django.conf import settings

        if getattr(settings, 'AUTH_USER_MODEL', None) != 'email_auth.User':
            yield "settings.AUTH_USER_MODEL should be 'email_auth.User'."

        AUTHENTICATION_BACKENDS = getattr(settings, 'AUTHENTICATION_BACKENDS', [])
        if 'allauth.account.auth_backends.AuthenticationBackend' not in AUTHENTICATION_BACKENDS:
            yield "settings.AUTHENTICATION_BACKENDS should contain 'allauth.account.auth_backends.AuthenticationBackend'."

        if 'sass_processor.finders.CssFinder' not in getattr(settings, 'STATICFILES_FINDERS', []):
            yield "settings.STATICFILES_FINDERS should contain 'sass_processor.finders.CssFinder'."

        if 'node_modules' not in dict(getattr(settings, 'STATICFILES_DIRS', [])).keys():
            yield "settings.STATICFILES_DIRS should contain ('node_modules', '/…/node_modules')."

        if '/node_modules/' not in getattr(settings, 'NODE_MODULES_URL', ''):
            yield "settings.NODE_MODULES_URL should start with a URL pointing onto /…/node_modules/."

        for template_engine in getattr(settings, 'TEMPLATES', []):
            if template_engine['BACKEND'] != 'django.template.backends.django.DjangoTemplates':
                continue
            context_processors = template_engine['OPTIONS'].get('context_processors', [])
            if 'shop.context_processors.customer' not in context_processors:
                yield "'shop.context_processors.customer' is missing in 'context_processors' of the default Django Template engine."
            if 'shop.context_processors.shop_settings' not in context_processors:
                yield "'shop.context_processors.shop_settings' is missing in 'context_processors' of the default Django Template engine."
        for template_engine in getattr(settings, 'TEMPLATES', []):
            if template_engine['BACKEND'] == 'post_office.template.backends.post_office.PostOfficeTemplates':
                break
        else:
            yield "In settings.TEMPLATES, the backend for 'post_office.template.backends.post_office.PostOfficeTemplates' is missing."

        if getattr(settings, 'POST_OFFICE', {}).get('TEMPLATE_ENGINE') != 'post_office':
            yield "settings.POST_OFFICE should contain {'TEMPLATE_ENGINE': 'post_office'}."

        for dir in getattr(settings, 'SASS_PROCESSOR_INCLUDE_DIRS', []):
            if '/node_modules' in dir:
                break
        else:
            yield "settings.SASS_PROCESSOR_INCLUDE_DIRS should include the folder '…/node_modules'."

        if getattr(settings, 'COERCE_DECIMAL_TO_STRING', None) is not True:
            yield "settings.COERCE_DECIMAL_TO_STRING should be set to 'True'."

        if getattr(settings, 'FSM_ADMIN_FORCE_PERMIT', None) is not True:
            yield "settings.FSM_ADMIN_FORCE_PERMIT should be set to 'True'."

        if getattr(settings, 'SERIALIZATION_MODULES', {}).get('json') != 'shop.money.serializers':
            yield "settings.SERIALIZATION_MODULES['json'] should be set to 'shop.money.serializers'."

        REST_FRAMEWORK = getattr(settings, 'REST_FRAMEWORK', {})
        if 'shop.rest.money.JSONRenderer' not in REST_FRAMEWORK.get('DEFAULT_RENDERER_CLASSES', []):
            yield "settings.REST_FRAMEWORK['DEFAULT_RENDERER_CLASSES'] should contain class 'shop.rest.money.JSONRenderer'."

        if 'django_filters.rest_framework.DjangoFilterBackend' not in REST_FRAMEWORK.get('DEFAULT_FILTER_BACKENDS', []):
            yield "settings.REST_FRAMEWORK['DEFAULT_FILTER_BACKENDS'] should contain class 'django_filters.rest_framework.DjangoFilterBackend'."

        if getattr(settings, 'REST_AUTH_SERIALIZERS', {}).get('LOGIN_SERIALIZER') != 'shop.serializers.auth.LoginSerializer':
            yield "settings.REST_AUTH_SERIALIZERS['LOGIN_SERIALIZER'] should be set to 'shop.serializers.auth.LoginSerializer'."

        if 'shop.cascade' not in getattr(settings, 'CMSPLUGIN_CASCADE_PLUGINS', []):
            yield "settings.CMSPLUGIN_CASCADE_PLUGINS should contain entry 'shop.cascade'."

        CMSPLUGIN_CASCADE = getattr(settings, 'CMSPLUGIN_CASCADE', {})
        if CMSPLUGIN_CASCADE.get('link_plugin_classes') != [
            'shop.cascade.plugin_base.CatalogLinkPluginBase',
            'cmsplugin_cascade.link.plugin_base.LinkElementMixin',
            'shop.cascade.plugin_base.CatalogLinkForm']:
            yield "settings.CMSPLUGIN_CASCADE['link_plugin_classes'] should contain special classes able to link onto products."
        if CMSPLUGIN_CASCADE.get('bootstrap4', {}).get('template_basedir') != 'angular-ui':
            yield "settings.CMSPLUGIN_CASCADE['bootstrap4']['template_basedir'] should be 'angular-ui'."
        if CMSPLUGIN_CASCADE.get('segmentation_mixins') != [
            ('shop.cascade.segmentation.EmulateCustomerModelMixin',
             'shop.cascade.segmentation.EmulateCustomerAdminMixin')]:
            yield "settings.CMSPLUGIN_CASCADE['segmentation_mixins'] should contain a special version handling the Customer model."

        if not isinstance(getattr(settings, 'SHOP_CART_MODIFIERS', None), (list, tuple)):
            yield "settings.SHOP_CART_MODIFIERS should contain a list with cart modifiers."

    def deserialize_to_placeholder(self, page, data, slot='Main Content'):
        from cms.utils.i18n import get_public_languages

        language = get_public_languages()[0]
        placeholder = page.placeholders.get(slot=slot)
        deserialize_to_placeholder(placeholder, data, language)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.auth import logout, get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.tokens import default_token_generator
from django.core.exceptions import NON_FIELD_ERRORS
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from rest_framework import status
from rest_framework.views import APIView
from rest_framework.exceptions import ValidationError
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import AllowAny
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer
from rest_framework.response import Response
from rest_auth.views import LoginView as OriginalLoginView, PasswordChangeView as OriginalPasswordChangeView

from shop.models.cart import CartModel
from shop.models.customer import CustomerModel
from shop.rest.renderers import CMSPageRenderer
from shop.serializers.auth import PasswordResetRequestSerializer, PasswordResetConfirmSerializer
from shop.signals import email_queued


class AuthFormsView(GenericAPIView):
    """
    Generic view to handle authentication related forms such as user registration
    """
    serializer_class = None
    form_class = None

    def post(self, request, *args, **kwargs):
        if request.customer.is_visitor:
            customer = CustomerModel.objects.get_or_create_from_request(request)
        else:
            customer = request.customer
        form_data = request.data.get(self.form_class.scope_prefix, {})
        form = self.form_class(data=form_data, instance=customer)
        if form.is_valid():
            form.save(request=request)
            response_data = {form.form_name: {
                'success_message': _("Successfully registered yourself."),
            }}
            return Response(response_data, status=status.HTTP_200_OK)
        errors = dict(form.errors)
        if 'email' in errors:
            errors.update({NON_FIELD_ERRORS: errors.pop('email')})
        return Response({form.form_name: errors}, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class LoginView(OriginalLoginView):
    form_name = 'login_form'

    def login(self):
        """
        Logs in as the given user, and moves the items from the current to the new cart.
        """
        try:
            anonymous_cart = CartModel.objects.get_from_request(self.request)
        except CartModel.DoesNotExist:
            anonymous_cart = None
        if self.request.customer.user.is_anonymous or self.request.customer.is_registered:
            previous_user = None
        else:
            previous_user = self.request.customer.user
        super(LoginView, self).login()  # this rotates the session_key
        if not self.serializer.data.get('stay_logged_in'):
            self.request.session.set_expiry(0)  # log out when the browser is closed
        authenticated_cart = CartModel.objects.get_from_request(self.request)
        if anonymous_cart:
            # an anonymous customer logged in, now merge his current cart with a cart,
            # which previously might have been created under his account.
            authenticated_cart.merge_with(anonymous_cart)
        if previous_user and previous_user.is_active is False and previous_user != self.request.user:
            # keep the database clean and remove this anonymous entity
            if previous_user.customer.orders.count() == 0:
                previous_user.delete()

    def post(self, request, *args, **kwargs):
        self.request = request
        form_data = request.data.get('form_data', {})
        self.serializer = self.get_serializer(data=form_data)
        if self.serializer.is_valid():
            self.login()
            return self.get_response()

        exc = ValidationError({self.form_name: self.serializer.errors})
        response = self.handle_exception(exc)
        self.response = self.finalize_response(request, response, *args, **kwargs)
        return self.response


class LogoutView(APIView):
    """
    Calls Django logout method and delete the auth Token assigned to the current User object.
    """
    permission_classes = (AllowAny,)
    form_name = 'logout_form'

    def post(self, request):
        try:
            request.user.auth_token.delete()
        except:
            pass
        logout(request)
        request.user = AnonymousUser()
        response_data = {self.form_name: {'success_message': _("Successfully logged out.")}}
        return Response(response_data)


class PasswordChangeView(OriginalPasswordChangeView):
    form_name = 'password_change_form'

    def post(self, request, *args, **kwargs):
        form_data = request.data.get('form_data', {})
        serializer = self.get_serializer(data=form_data)
        if serializer.is_valid():
            serializer.save()
            response_data = {self.form_name: {
                'success_message': _("Password has been changed successfully."),
            }}
            return Response(response_data)
        return Response({self.form_name: serializer.errors}, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class PasswordResetRequestView(GenericAPIView):
    """
    Calls Django Auth PasswordResetRequestForm save method.

    Accepts the following POST parameters: email
    Returns the success/fail message.
    """
    serializer_class = PasswordResetRequestSerializer
    permission_classes = (AllowAny,)
    form_name = 'password_reset_request_form'

    def post(self, request, *args, **kwargs):
        form_data = request.data.get('form_data', {})
        serializer = self.get_serializer(data=form_data)
        if not serializer.is_valid():
            return Response({self.form_name: serializer.errors}, status=status.HTTP_422_UNPROCESSABLE_ENTITY)

        # send email containing a reset link
        serializer.save()

        # trigger async email queue
        email_queued()

        # Return the success message with OK HTTP status
        msg = _("Instructions on how to reset the password have been sent to '{email}'.")
        response_data = {self.form_name: {
            'success_message': msg.format(**serializer.data),
        }}
        return Response(response_data)


class PasswordResetConfirmView(GenericAPIView):
    """
    Password reset e-mail link points onto a CMS page with the Page ID = 'password-reset-confirm'.
    This page then shall render the CMS plugin as provided by the **ShopAuthenticationPlugin** using
    the form "Confirm Password Reset".
    """
    renderer_classes = (CMSPageRenderer, JSONRenderer, BrowsableAPIRenderer)
    serializer_class = PasswordResetConfirmSerializer
    permission_classes = (AllowAny,)
    token_generator = default_token_generator
    form_name = 'password_reset_confirm_form'

    def get(self, request, uidb64=None, token=None):
        data = {'uid': uidb64, 'token': token}
        serializer_class = self.get_serializer_class()
        password = get_user_model().objects.make_random_password()
        data.update(new_password1=password, new_password2=password)
        serializer = serializer_class(data=data, context=self.get_serializer_context())
        if not serializer.is_valid():
            return Response({'validlink': False})
        return Response({
            'validlink': True,
            'user_name': force_text(serializer.user),
            'form_name': 'password_reset_form',
        })

    def post(self, request, uidb64=None, token=None):
        try:
            data = dict(request.data['form_data'], uid=uidb64, token=token)
        except (KeyError, TypeError, ValueError):
            errors = {'non_field_errors': [_("Invalid POST data.")]}
        else:
            serializer = self.get_serializer(data=data)
            if serializer.is_valid():
                serializer.save()
                response_data = {self.form_name: {
                    'success_message': _("Password has been reset with the new password."),
                }}
                return Response(response_data)
            else:
                errors = serializer.errors
        return Response({self.form_name: errors}, status=status.HTTP_422_UNPROCESSABLE_ENTITY)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import transaction

from rest_framework import status
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.renderers import BrowsableAPIRenderer

from shop.models.address import ShippingAddressModel, BillingAddressModel
from shop.models.cart import CartModel
from shop.rest.money import JSONRenderer


class AddressEditView(GenericAPIView):
    """
    View class to render associated addresses for the current user.
    """
    renderer_classes = (JSONRenderer, BrowsableAPIRenderer)
    permission_classes = (IsAuthenticated,)
    form_class = None  # must be overridde

    def __init__(self, **kwargs):
        super(AddressEditView, self).__init__(**kwargs)
        self.visible_fields = [f.name for f in self.form_class().visible_fields()]

    def get(self, request, priority=None, *args, **kwargs):
        if priority == 'add':
            # deliver an empty form
            form = self.form_class()
        else:
            try:
                if self.form_class.__name__ == 'BillingAddressForm':
                    instance = request.customer.billingaddress_set.get(priority=priority)
                elif self.form_class.__name__ == 'ShippingAddressForm':
                    instance = request.customer.shippingaddress_set.get(priority=priority)
                else:
                    raise CartModel.DoesNotExist()
            except (ShippingAddressModel.DoesNotExist, BillingAddressModel.DoesNotExist):
                return Response(status=status.HTTP_410_GONE)
            else:
                cart = CartModel.objects.get_from_request(request)
                form = self.form_class(instance=instance, cart=cart)
        initial_data = dict((k, v) for k, v in form.get_initial_data().items() if k in self.visible_fields)
        initial_data.pop('use_primary_address', None)
        return Response({self.form_class.form_name: initial_data})

    def put(self, request, *args, **kwargs):
        data = request.data.get(self.form_class.scope_prefix, {})
        cart = CartModel.objects.get_from_request(request)
        form = self.form_class(data=data, cart=cart)
        if form.is_valid():
            return Response()
        response_data = {self.form_class.form_name: dict(errors=form.errors)}
        return Response(response_data, status=status.HTTP_422_UNPROCESSABLE_ENTITY)

    def delete(self, request, priority=None, *args, **kwargs):
        cart = CartModel.objects.get_from_request(request)
        with transaction.atomic():
            try:
                if self.form_class.__name__ == 'BillingAddressForm':
                    request.customer.billingaddress_set.get(priority=priority).delete()
                elif self.form_class.__name__ == 'ShippingAddressForm':
                    request.customer.shippingaddress_set.get(priority=priority).delete()
            except (ValueError, ShippingAddressModel.DoesNotExist, BillingAddressModel.DoesNotExist):
                pass

            # take the last of the remaining addresses
            if self.form_class.__name__ == 'BillingAddressForm':
                instance = request.customer.billingaddress_set.last()
                cart.billing_address = instance
            elif self.form_class.__name__ == 'ShippingAddressForm':
                instance = request.customer.shippingaddress_set.last()
                cart.shipping_address = instance
            cart.save()

        # repopulate the form with address fields from the last remaining address
        if instance:
            form = self.form_class(instance=instance, cart=cart)
            initial_data = dict((k, v) for k, v in form.get_initial_data().items() if k in self.visible_fields)
            initial_data.update(active_priority=str(instance.priority), siblings_summary=form.siblings_summary)
            return Response({self.form_class.form_name: initial_data})
        return Response(status=status.HTTP_410_GONE)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rest_framework import serializers
from shop.conf import app_settings
from shop.models.cart import CartModel, CartItemModel
from shop.rest.money import MoneyField
from shop.models.fields import ChoiceEnum


class ExtraCartRow(serializers.Serializer):
    """
    This data structure holds extra information for each item, or for the whole cart, while
    processing the cart using their modifiers.
    """
    label = serializers.CharField(
        read_only=True,
        help_text="A short description of this row in a natural language.",
    )

    amount = MoneyField(
        help_text="The price difference, if applied.",
    )


class ExtraCartRowList(serializers.Serializer):
    """
    Represent the OrderedDict used for cart.extra_rows and cart_item.extra_rows.
    Additionally add the modifiers identifier to each element.
    """
    def to_representation(self, obj):
        return [dict(ecr.data, modifier=modifier) for modifier, ecr in obj.items()]


class BaseItemSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(lookup_field='pk', view_name='shop:cart-detail')
    unit_price = MoneyField()
    line_total = MoneyField()
    summary = serializers.SerializerMethodField(
        help_text="Sub-serializer for fields to be shown in the product's summary.")
    extra_rows = ExtraCartRowList(read_only=True)

    class Meta:
        model = CartItemModel

    def create(self, validated_data):
        assert 'cart' in validated_data
        cart_item = CartItemModel.objects.get_or_create(**validated_data)[0]
        cart_item.save()
        return cart_item

    def to_representation(self, cart_item):
        cart_item.update(self.context['request'])
        representation = super(BaseItemSerializer, self).to_representation(cart_item)
        return representation

    def validate_product(self, product):
        if not product.active:
            msg = "Product `{}` is inactive, and can not be added to the cart."
            raise serializers.ValidationError(msg.format(product))
        return product

    def get_summary(self, cart_item):
        serializer_class = app_settings.PRODUCT_SUMMARY_SERIALIZER
        serializer = serializer_class(cart_item.product, context=self.context,
                                      read_only=True, label=self.root.label)
        return serializer.data


class CartItemSerializer(BaseItemSerializer):
    class Meta(BaseItemSerializer.Meta):
        exclude = ['cart', 'id']

    def create(self, validated_data):
        validated_data['cart'] = CartModel.objects.get_or_create_from_request(self.context['request'])
        return super(CartItemSerializer, self).create(validated_data)


class WatchItemSerializer(BaseItemSerializer):
    class Meta(BaseItemSerializer.Meta):
        fields = ['product', 'product_code', 'url', 'summary', 'quantity', 'extra']

    def create(self, validated_data):
        cart = CartModel.objects.get_or_create_from_request(self.context['request'])
        validated_data.update(cart=cart, quantity=0)
        return super(WatchItemSerializer, self).create(validated_data)


class CartItems(ChoiceEnum):
    without = False
    unsorted = 1
    arranged = 2


class BaseCartSerializer(serializers.ModelSerializer):
    subtotal = MoneyField()
    total = MoneyField()
    extra_rows = ExtraCartRowList(read_only=True)

    class Meta:
        model = CartModel
        fields = ['subtotal', 'total', 'extra_rows']

    def to_representation(self, cart):
        cart.update(self.context['request'])
        representation = super(BaseCartSerializer, self).to_representation(cart)
        if self.with_items:
            items = self.represent_items(cart)
            representation.update(items=items)
        return representation

    def represent_items(self, cart):
        raise NotImplementedError("{} must implement method `represent_items()`.".format(self.__class__))


class CartSerializer(BaseCartSerializer):
    total_quantity = serializers.IntegerField()
    num_items = serializers.IntegerField()

    class Meta(BaseCartSerializer.Meta):
        fields = ['total_quantity', 'num_items'] + BaseCartSerializer.Meta.fields

    def __init__(self, *args, **kwargs):
        self.with_items = kwargs.pop('with_items', CartItems.without)
        super(CartSerializer, self).__init__(*args, **kwargs)

    def represent_items(self, cart):
        if self.with_items == CartItems.unsorted:
            items = CartItemModel.objects.filter(cart=cart, quantity__gt=0).order_by('-updated_at')
        else:
            items = CartItemModel.objects.filter_cart_items(cart, self.context['request'])
        serializer = CartItemSerializer(items, context=self.context, label=self.label, many=True)
        return serializer.data


class WatchSerializer(BaseCartSerializer):
    num_items = serializers.IntegerField()

    class Meta(BaseCartSerializer.Meta):
        fields = ['num_items']

    def __init__(self, *args, **kwargs):
        self.with_items = kwargs.pop('with_items', CartItems.arranged)
        super(WatchSerializer, self).__init__(*args, **kwargs)

    def represent_items(self, cart):
        if self.with_items == CartItems.unsorted:
            items = CartItemModel.objects.filter(cart=cart, quantity=0).order_by('-updated_at')
        else:
            items = CartItemModel.objects.filter_watch_items(cart, self.context['request'])
        serializer = WatchItemSerializer(items, context=self.context, label=self.label, many=True)
        return serializer.data



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.contrib.auth import get_user_model, authenticate, login, password_validation
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import EmailMultiAlternatives
from django.core.exceptions import ValidationError
from django.forms import widgets, ModelForm
from django.template.loader import get_template, select_template, render_to_string
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from djng.forms import fields, NgModelFormMixin, NgFormValidationMixin
from djng.styling.bootstrap3.forms import Bootstrap3ModelForm
from post_office import mail as post_office_mail
from post_office.models import EmailTemplate
from shop.conf import app_settings
from shop.forms.base import UniqueEmailValidationMixin
from shop.models.customer import CustomerModel
from shop.signals import email_queued


class RegisterUserForm(NgModelFormMixin, NgFormValidationMixin, UniqueEmailValidationMixin, Bootstrap3ModelForm):
    form_name = 'register_user_form'
    scope_prefix = 'form_data'
    field_css_classes = 'input-group has-feedback'

    email = fields.EmailField(
        label=_("Your e-mail address"),
        widget=widgets.EmailInput(attrs={'placeholder': _("E-mail address")})
    )

    preset_password = fields.BooleanField(
        label=_("Preset password"),
        widget=widgets.CheckboxInput(attrs={'class': 'form-check-input'}),
        required=False,
        help_text=_("Send a randomly generated password to your e-mail address."),
    )

    error_messages = {
        'password_mismatch': _("The two password fields didn't match."),
    }

    password1 = fields.CharField(
        label=_("New password"),
        widget=widgets.PasswordInput(attrs={'placeholder': _("Password")}),
        strip=False,
        help_text=password_validation.password_validators_help_text_html(),
    )

    password2 = fields.CharField(
        label=_("New password confirmation"),
        strip=False,
        widget=widgets.PasswordInput(attrs={'placeholder': _("Password")}),
        help_text=format_html('<ul><li>{}</li></ul>', _("Confirm the password.")),
    )

    class Meta:
        model = CustomerModel
        fields = ['email', 'password1', 'password2']

    def __init__(self, data=None, instance=None, *args, **kwargs):
        if data and data.get('preset_password', False):
            pwd_length = max(self.base_fields['password1'].min_length or 8, 8)
            password = get_user_model().objects.make_random_password(pwd_length)
            data['password1'] = data['password2'] = password
        super(RegisterUserForm, self).__init__(data=data, instance=instance, *args, **kwargs)

    def clean(self):
        cleaned_data = super(RegisterUserForm, self).clean()
        password1 = cleaned_data.get('password1')
        password2 = cleaned_data.get('password2')
        if password1 and password2:
            if password1 != password2:
                raise ValidationError(
                    self.error_messages['password_mismatch'],
                    code='password_mismatch',
                )
        password_validation.validate_password(password2)
        return cleaned_data

    def save(self, request=None, commit=True):
        self.instance.user.is_active = True
        self.instance.user.email = self.cleaned_data['email']
        self.instance.user.set_password(self.cleaned_data['password1'])
        self.instance.recognize_as_registered(request, commit=False)
        customer = super(RegisterUserForm, self).save(commit)
        password = self.cleaned_data['password1']
        if self.cleaned_data['preset_password']:
            self._send_password(request, customer.user, password)
        user = authenticate(username=customer.user.username, password=password)
        login(request, user)
        return customer

    def _send_password(self, request, user, password):
        current_site = get_current_site(request)
        context = {
            'site_name': current_site.name,
            'absolute_base_uri': request.build_absolute_uri('/'),
            'email': user.email,
            'password': password,
            'user': user,
        }
        subject_template = select_template([
            '{}/email/register-user-subject.txt'.format(app_settings.APP_LABEL),
            'shop/email/register-user-subject.txt',
        ])
        # Email subject *must not* contain newlines
        subject = ''.join(subject_template.render(context).splitlines())
        body_text_template = select_template([
            '{}/email/register-user-body.txt'.format(app_settings.APP_LABEL),
            'shop/email/register-user-body.txt',
        ])
        body_html_template = select_template([
            '{}/email/register-user-body.html'.format(app_settings.APP_LABEL),
            'shop/email/register-user-body.html',
        ], using='post_office')
        message = body_text_template.render(context)
        html_message = body_html_template.render(context)
        from_email = getattr(settings, 'DEFAULT_FROM_EMAIL')
        user.email_user(subject, message, from_email=from_email, html_message=html_message)
        email_queued()


class ContinueAsGuestForm(ModelForm):
    """
    Handles Customer's decision to order as guest.
    """
    form_name = 'continue_as_guest_form'
    scope_prefix = 'form_data'

    class Meta:
        model = CustomerModel
        fields = ()  # this form doesn't show any fields

    def save(self, request=None, commit=True):
        self.instance.recognize_as_guest(request, commit=False)
        self.instance.user.is_active = app_settings.GUEST_IS_ACTIVE_USER
        if self.instance.user.is_active:
            # set a usable password, otherwise the user later can not reset its password
            password = get_user_model().objects.make_random_password(length=30)
            self.instance.user.set_password(password)
        return super(ContinueAsGuestForm, self).save(commit)


class PasswordResetRequestForm(PasswordResetForm):
    def send_mail(self, subject_template_name, email_template_name,
                  context, from_email, to_email, html_email_template_name=None):
        try:
            email_template = EmailTemplate.objects.get(name='password-reset-inform')
        except EmailTemplate.DoesNotExist:
            subject = render_to_string(subject_template_name, context)
            # Email subject *must not* contain newlines
            subject = ''.join(subject.splitlines())
            body = render_to_string(email_template_name, context)

            email_message = EmailMultiAlternatives(subject, body, from_email, [to_email])
            if html_email_template_name:
                template = get_template(html_email_template_name, using='post_office')
                html = template.render(context)
                email_message.attach_alternative(html, 'text/html')
                template.attach_related(email_message)
            email_message.send()
        else:
            context['user'] = str(context['user'])
            context['uid'] = context['uid'].decode('utf-8')
            post_office_mail.send(to_email, template=email_template, context=context, render_on_delivery=True)
        email_queued()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rest_framework import serializers
from shop.models.product import ProductModel


class ProductSelectSerializer(serializers.ModelSerializer):
    """
    A simple serializer to convert the product's name and code used for rendering the
    `Select2 Widget`_'s content, while looking up for a certain product.
    This serializer shall return a list of 2-tuples, whose 1st entry is the
    primary key of the product and the second entry is the rendered name.

    .. _Select2 Widget: https://github.com/applegrew/django-select2
    """
    text = serializers.SerializerMethodField()

    class Meta:
        model = ProductModel
        fields = ['id', 'text']

    def get_text(self, instance):
        return instance.product_name



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _

from shop.models.address import BaseShippingAddress, BaseBillingAddress, CountryField


class AddressModelMixin(models.Model):
    name = models.CharField(
        _("Full name"),
        max_length=1024,
    )

    address1 = models.CharField(
        _("Address line 1"),
        max_length=1024,
    )

    address2 = models.CharField(
        _("Address line 2"),
        max_length=1024,
        blank=True,
        null=True,
    )

    zip_code = models.CharField(
        _("ZIP / Postal code"),
        max_length=12,
    )

    city = models.CharField(
        _("City"),
        max_length=1024,
    )

    country = CountryField(_("Country"))

    class Meta:
        abstract = True


class ShippingAddress(BaseShippingAddress, AddressModelMixin):
    class Meta:
        verbose_name = _("Shipping Address")
        verbose_name_plural = _("Shipping Addresses")


class BillingAddress(BaseBillingAddress, AddressModelMixin):
    class Meta:
        verbose_name = _("Billing Address")
        verbose_name_plural = _("Billing Addresses")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from shop.models.delivery import BaseDelivery


class Delivery(BaseDelivery):
    """Default materialized model for OrderShipping"""
    class Meta(BaseDelivery.Meta):
        verbose_name = _("Delivery")
        verbose_name_plural = _("Deliveries")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _
from django_fsm import transition, RETURN_VALUE
from shop.models.order import BaseOrder


class ManualPaymentWorkflowMixin(object):
    """
    Add this class to `settings.SHOP_ORDER_WORKFLOWS` to mix it into your `OrderModel`.
    It adds all the methods required for state transitions, when used with the
    `ForwardFundPayment` provider from above.
    """
    TRANSITION_TARGETS = {
        'awaiting_payment': _("Awaiting a forward fund payment"),
        'prepayment_deposited': _("Prepayment deposited"),
        'no_payment_required': _("No Payment Required"),
    }
    _manual_payment_transitions = TRANSITION_TARGETS.keys()

    def __init__(self, *args, **kwargs):
        if not isinstance(self, BaseOrder):
            raise ImproperlyConfigured("class 'ManualPaymentWorkflowMixin' is not of type 'BaseOrder'")
        CancelOrderWorkflowMixin.CANCELABLE_SOURCES.update(self._manual_payment_transitions)
        super(ManualPaymentWorkflowMixin, self).__init__(*args, **kwargs)

    @transition(field='status', source=['created'], target='no_payment_required')
    def no_payment_required(self):
        """
        Signals that an Order can proceed directly, by confirming a payment of value zero.
        """

    @transition(field='status', source=['created'], target='awaiting_payment')
    def awaiting_payment(self):
        """
        Signals that the current Order awaits a payment.
        Invoked by ForwardFundPayment.get_payment_request.
        """

    def payment_deposited(self):
        if hasattr(self, 'amount_paid'):
            del self.amount_paid
        return self.amount_paid > 0

    @transition(field='status', source=['awaiting_payment'],
                target=RETURN_VALUE('awaiting_payment', 'prepayment_deposited'),
                conditions=[payment_deposited],
                custom=dict(admin=True, button_name=_("Payment Received")))
    def prepayment_deposited(self):
        """
        Signals that the current Order received a payment.
        """
        return 'prepayment_deposited' if self.is_fully_paid() else 'awaiting_payment'

    @transition(field='status', source=['prepayment_deposited', 'no_payment_required'],
                custom=dict(auto=True))
    def acknowledge_prepayment(self):
        """
        Acknowledge the payment. This method is invoked automatically.
        """
        self.acknowledge_payment()

    @transition(field='status', source='refund_payment', target=RETURN_VALUE('refund_payment', 'order_canceled'),
                custom=dict(admin=True, button_name=_("Mark as Refunded")))
    def payment_refunded(self):
        """
        Signals that the payment for this Order has been refunded manually.
        """
        return 'refund_payment' if self.amount_paid else 'order_canceled'


class CancelOrderWorkflowMixin(object):
    """
    Add this class to `settings.SHOP_ORDER_WORKFLOWS` to mix it into your `OrderModel`.
    It adds all the methods required for state transitions, to cancel an order.
    """
    CANCELABLE_SOURCES = {'new', 'created', 'payment_confirmed', 'payment_declined', 'ready_for_delivery'}
    TRANSITION_TARGETS = {
        'refund_payment': _("Refund payment"),
        'order_canceled': _("Order Canceled"),
    }

    def cancelable(self):
        return super(CancelOrderWorkflowMixin, self).cancelable() or self.status in self.CANCELABLE_SOURCES

    @transition(field='status', target=RETURN_VALUE(*TRANSITION_TARGETS.keys()),
                conditions=[cancelable], custom=dict(admin=True, button_name=_("Cancel Order")))
    def cancel_order(self):
        """
        Signals that an Order shall be canceled.
        """
        self.withdraw_from_delivery()
        if self.amount_paid:
            self.refund_payment()
        return 'refund_payment' if self.amount_paid else 'order_canceled'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser
from django.db import models
from django.http.request import HttpRequest
from django.utils.six.moves.urllib.parse import urlparse
from post_office import mail
from post_office.models import EmailTemplate
from shop.conf import app_settings
from shop.models.order import BaseOrder
from shop.models.notification import Notification
from shop.serializers.delivery import DeliverySerializer
from shop.serializers.order import OrderDetailSerializer
from shop.signals import email_queued


class EmulateHttpRequest(HttpRequest):
    """
    Use this class to emulate a HttpRequest object, when templates must be rendered
    asynchronously, for instance when an email must be generated out of an Order object.
    """
    def __init__(self, customer, stored_request):
        super(EmulateHttpRequest, self).__init__()
        parsedurl = urlparse(stored_request.get('absolute_base_uri'))
        self.path = self.path_info = parsedurl.path
        self.environ = {}
        self.META['PATH_INFO'] = parsedurl.path
        self.META['SCRIPT_NAME'] = ''
        self.META['HTTP_HOST'] = parsedurl.netloc
        self.META['HTTP_X_FORWARDED_PROTO'] = parsedurl.scheme
        self.META['QUERY_STRING'] = parsedurl.query
        self.META['HTTP_USER_AGENT'] = stored_request.get('user_agent')
        self.META['REMOTE_ADDR'] = stored_request.get('remote_ip')
        self.method = 'GET'
        self.LANGUAGE_CODE = self.COOKIES['django_language'] = stored_request.get('language')
        self.customer = customer
        self.user = customer.is_anonymous and AnonymousUser or customer.user
        self.current_page = None


def transition_change_notification(order):
    """
    This function shall be called, after an Order object performed a transition change.

    """
    if not isinstance(order, BaseOrder):
        raise TypeError("Object order must inherit from class BaseOrder")
    emails_in_queue = False
    for notification in Notification.objects.filter(transition_target=order.status):
        recipient = notification.get_recipient(order)
        if recipient is None:
            continue

        # emulate a request object which behaves similar to that one, when the customer submitted its order
        emulated_request = EmulateHttpRequest(order.customer, order.stored_request)
        customer_serializer = app_settings.CUSTOMER_SERIALIZER(order.customer)
        render_context = {'request': emulated_request, 'render_label': 'email'}
        order_serializer = OrderDetailSerializer(order, context=render_context)
        language = order.stored_request.get('language')
        context = {
            'customer': customer_serializer.data,
            'order': order_serializer.data,
            'ABSOLUTE_BASE_URI': emulated_request.build_absolute_uri().rstrip('/'),
            'render_language': language,
        }
        try:
            latest_delivery = order.delivery_set.latest()
            context['latest_delivery'] = DeliverySerializer(latest_delivery, context=render_context).data
        except (AttributeError, models.ObjectDoesNotExist):
            pass
        try:
            template = notification.mail_template.translated_templates.get(language=language)
        except EmailTemplate.DoesNotExist:
            template = notification.mail_template
        attachments = {}
        for notiatt in notification.notificationattachment_set.all():
            attachments[notiatt.attachment.original_filename] = notiatt.attachment.file.file
        mail.send(recipient, template=template, context=context,
                  attachments=attachments, render_on_delivery=True)
        emails_in_queue = True
    if emails_in_queue:
        email_queued()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rest_framework import serializers
from shop.serializers.bases import ProductSerializer


class ProductSummarySerializer(ProductSerializer):
    """
    Default serializer to create a summary from our Product model. This summary then is used to
    render various list views, such as the catalog-, the cart-, and the list of ordered items.
    In case the Product model is polymorphic, this shall serialize the smallest common denominator
    of all product information.
    """
    media = serializers.SerializerMethodField(
        help_text="Returns a rendered HTML snippet containing a sample image among other elements",
    )

    caption = serializers.SerializerMethodField(
        help_text="Returns the content from caption field if available",
    )

    class Meta(ProductSerializer.Meta):
        fields = ['id', 'product_name', 'product_url', 'product_model', 'price', 'media', 'caption']

    def get_media(self, product):
        return self.render_html(product, 'media')

    def get_caption(self, product):
        return getattr(product, 'caption', None)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from shop.modifiers.base import BaseCartModifier


class ShippingModifier(BaseCartModifier):
    """
    Base class for all shipping modifiers.
    """
    def get_choice(self):
        """
        :returns: A tuple consisting of 'value, label' used by the shipping form dialog to render
        the available shipping choices.
        """
        raise NotImplemented("{} must implement method `get_choice()`.".format(self.__class__))

    def is_active(self, shipping_modifier):
        """
        :returns: ``True`` if this shipping modifier is active.
        """
        return shipping_modifier == self.identifier

    def is_disabled(self, cart):
        """
        Hook method to be overridden by the concrete shipping modifier. Shall be used to
        temporarily disable a shipping method, in case the cart does not fulfill certain criteria,
        for instance an undeliverable destination address.

        :returns: ``True`` if this shipping modifier is disabled for the current cart.
        """
        return False

    def update_render_context(self, context):
        """
        Hook to update the rendering context with shipping specific data.
        """
        from shop.models.cart import CartModel

        if 'shipping_modifiers' not in context:
            context['shipping_modifiers'] = {}
        try:
            cart = CartModel.objects.get_from_request(context['request'])
            if self.is_active(cart.extra.get('shipping_modifier')):
                cart.update(context['request'])
                data = cart.extra_rows[self.identifier].data
                data.update(modifier=self.identifier)
                context['shipping_modifiers']['initial_row'] = data
        except (KeyError, CartModel.DoesNotExist):
            pass

    def ship_the_goods(self, delivery):
        """
        Hook to be overridden by the active shipping modifier. It should be used to perform the
        shipping request.
        """
        delivery.shipped_at = timezone.now()


class SelfCollectionModifier(ShippingModifier):
    """
    This modifiers has not influence on the cart final. It can be used,
    to enable the customer to pick up the products in the shop.
    """
    identifier = 'self-collection'

    def get_choice(self):
        return (self.identifier, _("Self-collection"))

    def ship_the_goods(self, delivery):
        if not delivery.shipping_id:
            delivery.shipping_id = str(delivery.id)
        super(SelfCollectionModifier, self).ship_the_goods(delivery)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _, pgettext_lazy
from shop.models import order


class OrderItem(order.BaseOrderItem):
    """Default materialized model for OrderItem"""
    quantity = models.IntegerField(_("Ordered quantity"))

    class Meta:
        verbose_name = pgettext_lazy('order_models', "Ordered Item")
        verbose_name_plural = pgettext_lazy('order_models', "Ordered Items")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.forms import widgets


class CheckboxInput(widgets.CheckboxInput):
    template_name = 'djng/forms/widgets/bootstrap3/checkbox.html'


class RadioSelect(widgets.RadioSelect):
    template_name = 'djng/forms/widgets/bootstrap3/radio.html'


class Select(widgets.Select):
    template_name = 'djng/forms/widgets/bootstrap3/select.html'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils import timezone
from django.views.decorators.cache import never_cache
from django.utils.translation import ugettext_lazy as _
from rest_framework import generics, mixins
from rest_framework.exceptions import NotFound, MethodNotAllowed
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import BasePermission
from shop.rest.money import JSONRenderer
from shop.rest.renderers import CMSPageRenderer
from shop.serializers.order import OrderListSerializer, OrderDetailSerializer
from shop.models.order import OrderModel


class OrderPagination(LimitOffsetPagination):
    default_limit = 15
    template = 'shop/templatetags/paginator.html'


class OrderPermission(BasePermission):
    """
    Allow access to a given Order if the user is entitled to.
    """
    def has_permission(self, request, view):
        if view.many and request.customer.is_visitor:
            detail = _("Only signed in customers can view their list of orders.")
            raise PermissionDenied(detail=detail)
        return True

    def has_object_permission(self, request, view, order):
        if request.user.is_authenticated:
            return order.customer.pk == request.user.pk
        if order.secret and order.secret == view.kwargs.get('secret'):
            return True
        detail = _("This order does not belong to you.")
        raise PermissionDenied(detail=detail)


class OrderView(mixins.ListModelMixin, mixins.RetrieveModelMixin, mixins.UpdateModelMixin,
                generics.GenericAPIView):
    """
    Base View class to render the fulfilled orders for the current user.
    """
    renderer_classes = [CMSPageRenderer, JSONRenderer, BrowsableAPIRenderer]
    list_serializer_class = OrderListSerializer
    detail_serializer_class = OrderDetailSerializer
    pagination_class = OrderPagination
    permission_classes = [OrderPermission]
    lookup_field = lookup_url_kwarg = 'slug'
    many = True
    last_order_lapse = timezone.timedelta(minutes=15)

    def get_queryset(self):
        queryset = OrderModel.objects.all()
        if not self.request.customer.is_visitor:
            queryset = queryset.filter(customer=self.request.customer).order_by('-updated_at')
        return queryset

    def get_serializer_class(self):
        if self.many:
            return self.list_serializer_class
        return self.detail_serializer_class

    def get_renderer_context(self):
        renderer_context = super(OrderView, self).get_renderer_context()
        if self.request.accepted_renderer.format == 'html':
            renderer_context.update(many=self.many)
            if not self.many:
                # add an extra ance to the breadcrumb to show the order number
                renderer_context.update(
                    is_last_order = self.is_last(),
                    extra_ance=self.get_object().get_number(),
                )
        return renderer_context

    def is_last(self):
        """
        Returns ``True`` if the given order is considered as the last order for its customer.
        This information may be used to distinguish between a "thank you" and a normal detail view.
        """
        assert self.many is False, "This method can be called for detail views only"
        lapse = timezone.now() - self.last_order_lapse
        current_order = self.get_object()
        last_order = self.get_queryset().first()
        return current_order.id == last_order.id and current_order.created_at > lapse

    @property
    def allowed_methods(self):
        """Restrict method "POST" only on the detail view"""
        allowed_methods = self._allowed_methods()
        if self.many:
            allowed_methods.remove('POST')
        return allowed_methods

    @never_cache
    def get(self, request, *args, **kwargs):
        if self.many:
            return self.list(request, *args, **kwargs)
        return self.retrieve(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        if self.many:
            raise MethodNotAllowed("Method POST is not allowed on Order List View")
        self.update(request, *args, **kwargs)
        return self.retrieve(request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        try:
            return super(OrderView, self).list(request, *args, **kwargs)
        except OrderModel.DoesNotExist:
            raise NotFound("No orders have been found for the current user.")

    def retrieve(self, request, *args, **kwargs):
        try:
            return super(OrderView, self).retrieve(request, *args, **kwargs)
        except OrderModel.DoesNotExist:
            raise NotFound("No order has been found for the current user.")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from rest_framework import serializers
from shop.conf import app_settings
from shop.models.delivery import DeliveryModel, DeliveryItemModel
from shop.modifiers.pool import cart_modifiers_pool


class DeliveryItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = DeliveryItemModel
        exclude = ['id', 'delivery', 'item']

    def to_representation(self, instance):
        data = app_settings.ORDER_ITEM_SERIALIZER(instance.item, context=self.context).data
        data['ordered_quantity'] = data.pop('quantity', None)
        data.update(super(DeliveryItemSerializer, self).to_representation(instance))
        return data


class DeliverySerializer(serializers.ModelSerializer):
    items = DeliveryItemSerializer(
        many=True,
        read_only=True,
    )

    number = serializers.CharField(source='get_number')
    shipping_method = serializers.SerializerMethodField()

    class Meta:
        model = DeliveryModel
        exclude = ['id', 'order']

    def get_shipping_method(self, instance):
        for shipping_modifier in cart_modifiers_pool.get_shipping_modifiers():
            value, label = shipping_modifier.get_choice()
            if value == shipping_modifier.identifier:
                break
        else:
            value, label = instance.shipping_method, instance.shipping_method
        return {'value': value, 'label': label}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured
from shop.models.order import OrderModel


class PaymentProvider(object):
    """
    Base class for all Payment Service Providers.
    """
    @property
    def namespace(self):
        """
        Use a unique namespace for this payment provider. It is used to build the communication URLs
        exposed to an external payment service provider.
        """
        msg = "The attribute `namespace` must be implemented by the class `{}`"
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def get_urls(self):
        """
        Return a list of URL patterns for external communication with the payment service provider.
        """
        return []

    def get_payment_request(self, cart, request):
        """
        Build a JavaScript expression which is evaluated by the success handler on the page
        submitting the purchase command. When redirecting to another page, use:
        ```
        window.location.href="URL-of-other-page";
        ```
        since this expression is evaluated inside an AngularJS directive.
        """
        return 'alert("Please implement method `get_payment_request` in the Python class inheriting from `PaymentProvider`!");'


class ForwardFundPayment(PaymentProvider):
    """
    Provides a simple prepayment payment provider.
    """
    namespace = 'forward-fund-payment'

    def __init__(self):
        if (not callable(getattr(OrderModel, 'no_payment_required', None)) or
            not callable(getattr(OrderModel, 'awaiting_payment', None))):
            msg = "Missing methods in Order model. Add 'shop.payment.workflows.ManualPaymentWorkflowMixin' to SHOP_ORDER_WORKFLOWS."
            raise ImproperlyConfigured(msg)
        super(ForwardFundPayment, self).__init__()

    def get_payment_request(self, cart, request):
        order = OrderModel.objects.create_from_cart(cart, request)
        order.populate_from_cart(cart, request)
        if order.total == 0:
            order.no_payment_required()
        else:
            order.awaiting_payment()
        order.save(with_notification=True)
        return 'window.location.href="{}";'.format(order.get_absolute_url())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from adminsortable2.admin import SortableAdminMixin
from cms.admin.placeholderadmin import PlaceholderAdminMixin, FrontendEditableAdminMixin
from shop.admin.product import CMSPageAsCategoryMixin, CMSPageFilter
from shop.models.defaults.commodity import Commodity

if settings.USE_I18N:
    from parler.admin import TranslatableAdmin


    @admin.register(Commodity)
    class CommodityAdmin(SortableAdminMixin, TranslatableAdmin, FrontendEditableAdminMixin,
                         PlaceholderAdminMixin, CMSPageAsCategoryMixin, admin.ModelAdmin):
        fieldsets = [
            (None, {
                'fields': ['product_name', 'slug', 'caption']
            }),
            (_("Common Fields"), {
                'fields': ['product_code', ('unit_price', 'active'), 'show_breadcrumb', 'sample_image'],
            }),
        ]
        filter_horizontal = ['cms_pages']
        list_filter = [CMSPageFilter]

        def get_prepopulated_fields(self, request, obj=None):
            return {
                'slug': ('product_name',)
            }

else:

    @admin.register(Commodity)
    class CommodityAdmin(SortableAdminMixin, FrontendEditableAdminMixin, PlaceholderAdminMixin,
                         CMSPageAsCategoryMixin, admin.ModelAdmin):
        fields = ['product_name', 'slug',  'caption', 'product_code',
                  ('unit_price', 'active'), 'show_breadcrumb', 'sample_image']
        filter_horizontal = ['cms_pages']
        prepopulated_fields = {'slug': ['product_name']}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from djangocms_text_ckeditor.models import Text
from djangocms_text_ckeditor.utils import plugin_tags_to_id_list, replace_plugin_tags
from cms.api import add_plugin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.models import CascadeElement


def deserialize_to_placeholder(placeholder, data, language):
    def plugins_from_data(placeholder, parent, data):
        for plugin_type, data, children_data in data:
            try:
                plugin_class = plugin_pool.get_plugin(plugin_type)
            except Exception:
                continue
            kwargs = dict(data)
            inlines = kwargs.pop('inlines', [])
            shared_glossary = kwargs.pop('shared_glossary', None)
            try:
                instance = add_plugin(placeholder, plugin_class, language, target=parent, **kwargs)
            except Exception:
                continue
            if isinstance(instance, CascadeElement):
                instance.plugin_class.add_inline_elements(instance, inlines)
                instance.plugin_class.add_shared_reference(instance, shared_glossary)

            # for some unknown reasons add_plugin sets instance.numchild to 0,
            # but fixing and save()-ing 'instance' executes some filters in an unwanted manner
            plugins_from_data(placeholder, instance, children_data)

            if isinstance(instance, Text):
                # we must convert the old plugin IDs into the new ones,
                # otherwise links are not displayed
                id_dict = dict(zip(
                    plugin_tags_to_id_list(instance.body),
                    (t[0] for t in instance.get_children().values_list('id'))
                ))
                instance.body = replace_plugin_tags(instance.body, id_dict)
                instance.save()

    plugins_from_data(placeholder, None, data['plugins'])



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-02-25 19:51
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('shop', '0008_notification_recipient'),
    ]

    operations = [
        migrations.DeleteModel(
            name='Email',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _

from shop.modifiers.base import BaseCartModifier
from shop.payment.providers import PaymentProvider, ForwardFundPayment


class PaymentModifier(BaseCartModifier):
    """
    Base class for all payment modifiers.
    """
    def __init__(self):
        assert isinstance(getattr(self, 'payment_provider', None), PaymentProvider), \
            "Each Payment modifier class requires a Payment Provider"
        super(PaymentModifier, self).__init__()

    @property
    def identifier(self):
        """
        Default identifier for payment providers.
        """
        return self.payment_provider.namespace

    def get_choice(self):
        """
        :returns: A tuple consisting of 'value, label' used by the payment form dialog to render
        the available payment choices.
        """
        raise NotImplemented("{} must implement method `get_choice()`.".format(self.__class__))

    def is_active(self, payment_modifier):
        """
        :returns: ``True`` if this payment modifier is active.
        """
        assert hasattr(self, 'payment_provider'), "A Payment Modifier requires a Payment Provider"
        return payment_modifier == self.payment_provider.namespace

    def is_disabled(self, cart):
        """
        Hook method to be overridden by the concrete payment modifier. Shall be used to
        temporarily disable a payment method, in case the cart does not fulfill certain criteria,
        for instance a too small total.

        :returns: ``True`` if this payment modifier is disabled for the current cart.
        """
        return False

    def update_render_context(self, context):
        """
        Hook to update the rendering context with payment specific data.
        """
        from shop.models.cart import CartModel

        if 'payment_modifiers' not in context:
            context['payment_modifiers'] = {}
        try:
            cart = CartModel.objects.get_from_request(context['request'])
            if self.is_active(cart.extra.get('payment_modifier')):
                cart.update(context['request'])
                data = cart.extra_rows[self.identifier].data
                data.update(modifier=self.identifier)
                context['payment_modifiers']['initial_row'] = data
        except (KeyError, CartModel.DoesNotExist):
            pass


class PayInAdvanceModifier(PaymentModifier):
    """
    This modifiers has no influence on the cart final. It can be used,
    to enable the customer to pay the products on delivery.
    """
    payment_provider = ForwardFundPayment()

    def get_choice(self):
        return (self.payment_provider.namespace, _("Pay in advance"))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.cache import add_never_cache_headers
from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from shop.models.cart import CartModel, CartItemModel
from shop.serializers.cart import CartSerializer, CartItemSerializer, WatchSerializer, WatchItemSerializer, CartItems


class BaseViewSet(viewsets.ModelViewSet):
    pagination_class = None
    with_items = CartItems.arranged

    def get_queryset(self):
        try:
            cart = CartModel.objects.get_from_request(self.request)
            if self.kwargs.get(self.lookup_field):
                # we're interest only into a certain cart item
                return CartItemModel.objects.filter(cart=cart)
            return cart
        except CartModel.DoesNotExist:
            return CartModel()

    def list(self, request, *args, **kwargs):
        cart = self.get_queryset()
        context = self.get_serializer_context()
        serializer = self.serializer_class(cart, context=context, label=self.serializer_label,
                                           with_items=self.with_items)
        return Response(serializer.data)

    def create(self, request, *args, **kwargs):
        """
        Create a new item in the cart.
        """
        context = self.get_serializer_context()
        item_serializer = self.item_serializer_class(context=context, data=request.data, label=self.serializer_label)
        item_serializer.is_valid(raise_exception=True)
        self.perform_create(item_serializer)
        headers = self.get_success_headers(item_serializer.data)
        return Response(item_serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def update(self, request, *args, **kwargs):
        """
        Handle changing the amount of the cart item referred by its primary key.
        """
        cart_item = self.get_object()
        context = self.get_serializer_context()
        item_serializer = self.item_serializer_class(cart_item, context=context, data=request.data,
                                                     label=self.serializer_label)
        item_serializer.is_valid(raise_exception=True)
        self.perform_update(item_serializer)
        cart_serializer = CartSerializer(cart_item.cart, context=context, label='cart')
        response_data = {
            'cart_item': item_serializer.data,
            'cart': cart_serializer.data,
        }
        return Response(data=response_data)

    def destroy(self, request, *args, **kwargs):
        """
        Delete a cart item referred by its primary key.
        """
        cart_item = self.get_object()
        context = self.get_serializer_context()
        cart_serializer = CartSerializer(cart_item.cart, context=context, label=self.serializer_label)
        self.perform_destroy(cart_item)
        response_data = {
            'cart_item': None,
            'cart': cart_serializer.data,
        }
        return Response(data=response_data)

    def finalize_response(self, request, response, *args, **kwargs):
        """Set HTTP headers to not cache this view"""
        if self.action != 'render_product_summary':
            add_never_cache_headers(response)
        return super(BaseViewSet, self).finalize_response(request, response, *args, **kwargs)


class CartViewSet(BaseViewSet):
    serializer_label = 'cart'
    serializer_class = CartSerializer
    item_serializer_class = CartItemSerializer

    @action(detail=True, methods=['get'])
    def fetch(self, request):
        cart = self.get_queryset()
        context = self.get_serializer_context()
        serializer = self.serializer_class(cart, context=context, with_items=CartItems.without)
        return Response(serializer.data)

    @action(detail=False, methods=['get'], url_path='fetch-dropdown')
    def fetch_dropdown(self, request):
        cart = self.get_queryset()
        context = self.get_serializer_context()
        serializer = self.serializer_class(cart, context=context, label='dropdown', with_items=CartItems.unsorted)
        return Response(serializer.data)


class WatchViewSet(BaseViewSet):
    serializer_label = 'watch'
    serializer_class = WatchSerializer
    item_serializer_class = WatchItemSerializer



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf.urls import url
from django.contrib import admin
from django.core.urlresolvers import reverse, NoReverseMatch
from django.db.models.fields import Field, FieldDoesNotExist
from django.forms import widgets
from django.http import HttpResponse
from django.template.loader import select_template
from django.utils.html import format_html
from django.utils.translation import pgettext_lazy
from fsm_admin.mixins import FSMTransitionMixin
from shop.conf import app_settings
from shop.models.customer import CustomerModel
from shop.models.order import OrderItemModel, OrderPayment
from shop.modifiers.pool import cart_modifiers_pool
from shop.serializers.order import OrderDetailSerializer
from shop.transition import transition_change_notification


class OrderPaymentInline(admin.TabularInline):
    model = OrderPayment
    extra = 0

    def get_formset(self, request, obj=None, **kwargs):
        """
        Convert the field `payment_method` into a select box with all possible payment methods.
        """
        choices = [pm.get_choice() for pm in cart_modifiers_pool.get_payment_modifiers()]
        kwargs.update(widgets={'payment_method': widgets.Select(choices=choices)})
        formset = super(OrderPaymentInline, self).get_formset(request, obj, **kwargs)
        return formset

    def has_add_permission(self, request, obj=None):
        assert obj is not None, "An Order object can not be added through the Django-Admin"
        return obj.status in ['awaiting_payment', 'refund_payment']

    def has_delete_permission(self, request, obj=None):
        return False

    def get_max_num(self, request, obj=None, **kwargs):
        qs = self.model.objects.filter(order=obj)
        if self.has_add_permission(request, obj):
            return qs.count() + 1
        return qs.count()

    def get_fields(self, request, obj=None):
        if self.has_add_permission(request, obj):
            return ['amount', 'transaction_id', 'payment_method', 'created_at']
        return ['get_amount', 'transaction_id', 'payment_method', 'created_at']

    def get_readonly_fields(self, request, obj=None):
        if self.has_add_permission(request, obj):
            return ['created_at', 'get_amount']
        # unless we expect a payment, set all fields as readonly
        return self.get_fields(request, obj)

    def get_amount(self, obj):
        """Return amount using correct local format"""
        return obj.amount
    get_amount.short_description = pgettext_lazy('admin', "Amount Paid")


class OrderItemInline(admin.StackedInline):
    model = OrderItemModel
    extra = 0
    fields = [
        ('product_code', 'unit_price', 'line_total',),
        ('quantity',),
        'render_as_html_extra',
    ]
    readonly_fields = ['product_code', 'quantity', 'unit_price', 'line_total', 'render_as_html_extra']
    template = 'shop/admin/edit_inline/stacked-order.html'

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_max_num(self, request, obj=None, **kwargs):
        return self.model.objects.filter(order=obj).count()

    def render_as_html_extra(self, obj):
        item_extra_template = select_template([
            '{0}/admin/orderitem-{1}-extra.html'.format(app_settings.APP_LABEL, obj.product.product_model),
            '{0}/admin/orderitem-product-extra.html'.format(app_settings.APP_LABEL),
            'shop/admin/orderitem-product-extra.html',
        ])
        return item_extra_template.render(obj.extra)
    render_as_html_extra.short_description = pgettext_lazy('admin', "Extra data")


class StatusListFilter(admin.SimpleListFilter):
    title = pgettext_lazy('admin', "Status")
    parameter_name = 'status'

    def lookups(self, request, model_admin):
        lookups = dict(model_admin.model._transition_targets)
        lookups.pop('new')
        lookups.pop('created')
        return lookups.items()

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(status=self.value())
        return queryset


class BaseOrderAdmin(FSMTransitionMixin, admin.ModelAdmin):
    list_display = ['get_number', 'customer', 'status_name', 'get_total', 'created_at']
    list_filter = [StatusListFilter]
    fsm_field = ['status']
    date_hierarchy = 'created_at'
    inlines = [OrderItemInline]
    readonly_fields = ['get_number', 'status_name', 'get_total', 'get_subtotal',
                       'get_customer_link', 'get_outstanding_amount', 'created_at', 'updated_at',
                       'render_as_html_extra', 'stored_request', 'is_fully_paid']
    fields = ['get_number', 'status_name',
              ('created_at', 'updated_at'),
              'get_customer_link',
              ('get_subtotal', 'get_total', 'get_outstanding_amount', 'is_fully_paid'),
              'render_as_html_extra', 'stored_request']
    actions = None
    change_form_template = 'shop/admin/change_form.html'

    def __init__(self, *args, **kwargs):
        super(BaseOrderAdmin, self).__init__(*args, **kwargs)
        self.extra_template = select_template([
            '{}/admin/order-extra.html'.format(app_settings.APP_LABEL),
            'shop/admin/order-extra.html',
        ])

    def get_number(self, obj):
        return obj.get_number()
    get_number.short_description = pgettext_lazy('admin', "Order number")

    def get_total(self, obj):
        return str(obj.total)
    get_total.short_description = pgettext_lazy('admin', "Total")

    def get_subtotal(self, obj):
        return str(obj.subtotal)
    get_subtotal.short_description = pgettext_lazy('admin', "Subtotal")

    def get_outstanding_amount(self, obj):
        return str(obj.outstanding_amount)
    get_outstanding_amount.short_description = pgettext_lazy('admin', "Outstanding amount")

    def is_fully_paid(self, obj):
        return obj.is_fully_paid()
    is_fully_paid.short_description = pgettext_lazy('admin', "Is fully paid")
    is_fully_paid.boolean = True

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def render_as_html_extra(self, obj):
        return self.extra_template.render(obj.extra)
    render_as_html_extra.short_description = pgettext_lazy('admin', "Extra data")

    def get_customer_link(self, obj):
        try:
            url = reverse('admin:shop_customerproxy_change', args=(obj.customer.pk,))
            return format_html('<a href="{0}" target="_new">{1}</a>', url, obj.customer.get_username())
        except NoReverseMatch:
            return format_html('<strong>{0}</strong>', obj.customer.get_username())
    get_customer_link.short_description = pgettext_lazy('admin', "Customer")

    def get_search_fields(self, request):
        search_fields = list(super(BaseOrderAdmin, self).get_search_fields(request))
        search_fields.extend(['customer__user__email', 'customer__user__last_name'])
        try:
            # if CustomerModel contains a number field, let search for it
            if isinstance(CustomerModel._meta.get_field('number'), Field):
                search_fields.append('customer__number')
        except FieldDoesNotExist:
            pass
        return search_fields

    def get_form(self, request, obj=None, **kwargs):
        ModelForm = super(BaseOrderAdmin, self).get_form(request, obj, **kwargs)
        if obj:
            # store the requested transition inside the instance, so that the model's `clean()` method can access it
            obj._fsm_requested_transition = self._get_requested_transition(request)
        return ModelForm

    def change_view(self, request, object_id, form_url='', extra_context=None):
        assert object_id, "An Order object can not be added through the Django-Admin"
        current_status = self.get_object(request, object_id).status
        response = super(BaseOrderAdmin, self).change_view(request, object_id, form_url, extra_context)
        order = self.get_object(request, object_id)
        if current_status != order.status:
            transition_change_notification(order)
        return response


class PrintInvoiceAdminMixin(object):
    """
    A customized OrderAdmin class shall inherit from this mixin class, to add
    methods for printing the the invoice.
    """
    def get_fields(self, request, obj=None):
        fields = list(super(PrintInvoiceAdminMixin, self).get_fields(request, obj))
        fields.append('print_out')
        return fields

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super(PrintInvoiceAdminMixin, self).get_readonly_fields(request, obj))
        readonly_fields.append('print_out')
        return readonly_fields

    def get_urls(self):
        my_urls = [
            url(r'^(?P<pk>\d+)/print_invoice/$', self.admin_site.admin_view(self.render_invoice),
                name='print_invoice'),
        ]
        my_urls.extend(super(PrintInvoiceAdminMixin, self).get_urls())
        return my_urls

    def _render_content(self, request, pk, template):
        order = self.get_object(request, pk)
        context = {'request': request, 'render_label': 'print'}
        customer_serializer = app_settings.CUSTOMER_SERIALIZER(order.customer)
        order_serializer = OrderDetailSerializer(order, context=context)
        return template.render(context={
            'customer': customer_serializer.data,
            'order': order_serializer.data,
            'object': order,
        }, request=request)

    def render_invoice(self, request, pk=None):
        template = select_template([
            '{}/print/invoice.html'.format(app_settings.APP_LABEL.lower()),
            'shop/print/invoice.html'
        ])
        content = self._render_content(request, pk, template)
        return HttpResponse(content)

    def print_out(self, obj):
        if obj.status in ['ready_for_delivery']:
            link = reverse('admin:print_invoice', args=(obj.id,)), pgettext_lazy('admin', "Invoice")
            return format_html(
                '<span class="object-tools"><a href="{0}" class="viewsitelink" target="_new">{1}</a></span>',
                *link)
        return ''
    print_out.short_description = pgettext_lazy('admin', "Print out")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from cms.models import Page
from cms.templatetags.cms_tags import PageAttribute as BrokenPageAttribute


class PageAttribute(BrokenPageAttribute):
    """
    This monkey patch can be withdrawn, after https://github.com/divio/django-cms/issues/5930
    has been fixed.
    """
    def get_value_for_context(self, context, **kwargs):
        try:
            return super(PageAttribute, self).get_value_for_context(context, **kwargs)
        except Page.DoesNotExist:
            return ''



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from decimal import Decimal
from shop.money import AbstractMoney, Money
from shop.modifiers.base import BaseCartModifier


class DefaultCartModifier(BaseCartModifier):
    """
    This modifier is required for almost every shopping cart. It handles the most basic
    calculations, ie. multiplying the items unit prices with the chosen quantity.
    Since this modifier sets the cart items line total, it must be listed as the first
    entry in `SHOP_CART_MODIFIERS`.
    """
    identifier = 'default'

    def process_cart_item(self, cart_item, request):
        cart_item.unit_price = cart_item.product.get_price(request)
        cart_item.line_total = cart_item.unit_price * cart_item.quantity
        return super(DefaultCartModifier, self).process_cart_item(cart_item, request)

    def process_cart(self, cart, request):
        if not isinstance(cart.subtotal, AbstractMoney):
            # if we don't know the currency, use the default
            cart.subtotal = Money(cart.subtotal)
        cart.total = cart.subtotal
        return super(DefaultCartModifier, self).process_cart(cart, request)


class WeightedCartModifier(BaseCartModifier):
    """
    This modifier is required for all shopping cart where we are interested into its weight.
    It sums up the weight of all articles, ie. multiplying the items weight with the chosen
    quantity.
    If this modifier is used, the classes implementing the product shall override their
    method ``get_weight()``, which must return the weight in kg as Decimal type.
    """
    identifier = 'weights'
    initial_weight = Decimal(0.01)  # in kg

    def pre_process_cart(self, cart, request):
        cart.weight = self.initial_weight
        return super(WeightedCartModifier, self).process_cart(cart, request)

    def pre_process_cart_item(self, cart, cart_item, request):
        cart.weight += Decimal(cart_item.product.get_weight() * cart_item.quantity)
        return super(WeightedCartModifier, self).process_cart_item(cart_item, request)
